# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import hashlib
import logging
import os
import shutil

from pex import pex_warnings
from pex.atomic_directory import atomic_directory
from pex.common import (
    Chroot,
    chmod_plus_x,
    filter_pyc_files,
    is_pyc_temporary_file,
    safe_copy,
    safe_delete,
    safe_mkdir,
    safe_mkdtemp,
    safe_open,
)
from pex.compatibility import commonpath, to_bytes
from pex.compiler import Compiler
from pex.dist_metadata import Distribution, MetadataError
from pex.enum import Enum
from pex.environment import PEXEnvironment
from pex.finders import get_entry_point_from_console_script, get_script_from_distributions
from pex.interpreter import PythonInterpreter
from pex.layout import Layout
from pex.orderedset import OrderedSet
from pex.pex import PEX
from pex.pex_info import PexInfo
from pex.sh_boot import create_sh_boot_script
from pex.targets import Targets
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING
from pex.util import CacheHelper

if TYPE_CHECKING:
    from typing import Dict, Optional


class CopyMode(Enum["CopyMode.Value"]):
    class Value(Enum.Value):
        pass

    COPY = Value("copy")
    LINK = Value("link")
    SYMLINK = Value("symlink")


BOOTSTRAP_ENVIRONMENT = """\
import os
import sys


__INSTALLED_FROM__ = '__PEX_EXE__'


def __re_exec__(argv0, *extra_launch_args):
  os.execv(argv0, [argv0] + list(extra_launch_args) + sys.argv[1:])


__execute__ = __name__ == "__main__"

def __maybe_install_pex__(pex, pex_root, pex_hash):
  from pex.layout import maybe_install
  from pex.tracer import TRACER

  installed_location = maybe_install(pex, pex_root, pex_hash)
  if not __execute__ or not installed_location:
    return installed_location

  # N.B.: This is read upon re-exec below to point sys.argv[0] back to the original pex before
  # unconditionally scrubbing the env var and handing off to user code.
  os.environ[__INSTALLED_FROM__] = pex

  TRACER.log('Executing installed PEX for {{}} at {{}}'.format(pex, installed_location))
  __re_exec__(sys.executable, installed_location)


def __maybe_run_venv__(pex, pex_root, pex_path):
  from pex.common import is_exe
  from pex.tracer import TRACER
  from pex.variables import venv_dir

  venv_dir = venv_dir(
    pex_file=pex,
    pex_root=pex_root, 
    pex_hash={pex_hash!r},
    has_interpreter_constraints={has_interpreter_constraints!r},
    pex_path=pex_path,
  )
  venv_pex = os.path.join(venv_dir, 'pex')
  if not __execute__ or not is_exe(venv_pex):
    # Code in bootstrap_pex will (re)create the venv after selecting the correct interpreter. 
    return venv_dir

  TRACER.log('Executing venv PEX for {{}} at {{}}'.format(pex, venv_pex))
  venv_python = os.path.join(venv_dir, 'bin', 'python')
  if {hermetic_venv_scripts!r}:
    __re_exec__(venv_python, '-sE', venv_pex)
  else:
    __re_exec__(venv_python, venv_pex)


def __entry_point_from_filename__(filename):
    # Either the entry point is "__main__" and we're in execute mode or "__pex__/__init__.py" and
    # we're in import hook mode.
    entry_point = os.path.dirname(filename)
    if __execute__:
        return entry_point
    return os.path.dirname(entry_point)


__entry_point__ = None
if '__file__' in locals() and __file__ is not None and os.path.exists(__file__):
  __entry_point__ = __entry_point_from_filename__(__file__)
elif '__loader__' in locals():
  from pkgutil import ImpLoader
  if hasattr(__loader__, 'archive'):
    __entry_point__ = __loader__.archive
  elif isinstance(__loader__, ImpLoader):
    __entry_point__ = __entry_point_from_filename__(__loader__.get_filename())

if __entry_point__ is None:
  sys.stderr.write('Could not launch python executable!\\n')
  sys.exit(2)

__installed_from__ = os.environ.pop(__INSTALLED_FROM__, None)
sys.argv[0] = __installed_from__ or sys.argv[0]

sys.path[0] = os.path.abspath(sys.path[0])
sys.path.insert(0, os.path.abspath(os.path.join(__entry_point__, {bootstrap_dir!r})))

__venv_dir__ = None
if not __installed_from__:
    os.environ['PEX'] = os.path.realpath(__entry_point__)
    from pex.variables import ENV, Variables
    __pex_root__ = Variables.PEX_ROOT.value_or(ENV, {pex_root!r})
    if not ENV.PEX_TOOLS and Variables.PEX_VENV.value_or(ENV, {is_venv!r}):
      __venv_dir__ = __maybe_run_venv__(
        __entry_point__,
        pex_root=__pex_root__,
        pex_path=ENV.PEX_PATH or {pex_path!r},
      )
    __installed_location__ = __maybe_install_pex__(
      __entry_point__, pex_root=__pex_root__, pex_hash={pex_hash!r}
    )
    if __installed_location__:
      __entry_point__ = __installed_location__
else:
    os.environ['PEX'] = os.path.realpath(__installed_from__)

from pex.pex_bootstrapper import bootstrap_pex
bootstrap_pex(__entry_point__, execute=__execute__, venv_dir=__venv_dir__)
"""


class PEXBuilder(object):
    """Helper for building PEX environments."""

    class Error(Exception):
        pass

    class ImmutablePEX(Error):
        pass

    class InvalidDistribution(Error):
        pass

    class InvalidDependency(Error):
        pass

    class InvalidExecutableSpecification(Error):
        pass

    def __init__(
        self,
        path=None,  # type: Optional[str]
        interpreter=None,  # type: Optional[PythonInterpreter]
        chroot=None,  # type: Optional[Chroot]
        pex_info=None,  # type: Optional[PexInfo]
        preamble=None,  # type: Optional[str]
        copy_mode=CopyMode.LINK,  # type: CopyMode.Value
    ):
        # type: (...) -> None
        """Initialize a pex builder.

        :keyword path: The path to write the PEX as it is built.  If ``None`` is specified,
          a temporary directory will be created.
        :keyword interpreter: The interpreter to use to build this PEX environment.  If ``None``
          is specified, the current interpreter is used.
        :keyword chroot: If specified, preexisting :class:`Chroot` to use for building the PEX.
        :keyword pex_info: A preexisting PexInfo to use to build the PEX.
        :keyword preamble: If supplied, execute this code prior to bootstrapping this PEX
          environment.
        :keyword copy_mode: Create the pex environment using the given copy mode.

        .. versionchanged:: 0.8
          The temporary directory created when ``path`` is not specified is now garbage collected on
          interpreter exit.
        """
        self._interpreter = interpreter or PythonInterpreter.get()
        self._chroot = chroot or Chroot(path or safe_mkdtemp())
        self._pex_info = pex_info or PexInfo.default()
        self._preamble = preamble or ""
        self._copy_mode = copy_mode

        self._shebang = self._interpreter.identity.hashbang()
        self._header = None  # type: Optional[str]
        self._logger = logging.getLogger(__name__)
        self._frozen = False
        self._distributions = {}  # type: Dict[str, Distribution]

    def _ensure_unfrozen(self, name="Operation"):
        if self._frozen:
            raise self.ImmutablePEX("%s is not allowed on a frozen PEX!" % name)

    @property
    def interpreter(self):
        return self._interpreter

    def chroot(self):
        # type: () -> Chroot
        return self._chroot

    def clone(self, into=None):
        """Clone this PEX environment into a new PEXBuilder.

        :keyword into: (optional) An optional destination directory to clone this PEXBuilder into.  If
          not specified, a temporary directory will be created.

        Clones PEXBuilder into a new location.  This is useful if the PEXBuilder has been frozen and
        rendered immutable.

        .. versionchanged:: 0.8
          The temporary directory created when ``into`` is not specified is now garbage collected on
          interpreter exit.
        """
        chroot_clone = self._chroot.clone(into=into)
        clone = self.__class__(
            chroot=chroot_clone,
            interpreter=self._interpreter,
            pex_info=self._pex_info.copy(),
            preamble=self._preamble,
            copy_mode=self._copy_mode,
        )
        clone.set_shebang(self._shebang)
        clone._distributions = self._distributions.copy()
        return clone

    def path(self):
        # type: () -> str
        return self.chroot().path()

    @property
    def info(self):
        return self._pex_info

    @info.setter
    def info(self, value):
        if not isinstance(value, PexInfo):
            raise TypeError("PEXBuilder.info must be a PexInfo!")
        self._ensure_unfrozen("Changing PexInfo")
        self._pex_info = value

    def add_source(self, filename, env_filename):
        """Add a source to the PEX environment.

        :param filename: The source filename to add to the PEX; None to create an empty file at
          `env_filename`.
        :param env_filename: The destination filename in the PEX.  This path
          must be a relative path.
        """
        self._ensure_unfrozen("Adding source")
        self._copy_or_link(filename, env_filename, "source")

    def add_resource(self, filename, env_filename):
        """Add a resource to the PEX environment.

        :param filename: The source filename to add to the PEX; None to create an empty file at
          `env_filename`.
        :param env_filename: The destination filename in the PEX.  This path
          must be a relative path.
        """
        pex_warnings.warn(
            "The `add_resource` method is deprecated. Resources should be added via the "
            "`add_source` method instead."
        )
        self._ensure_unfrozen("Adding a resource")
        self._copy_or_link(filename, env_filename, "resource")

    def add_requirement(self, req):
        """Add a requirement to the PEX environment.

        :param req: A requirement that should be resolved in this environment.

        .. versionchanged:: 0.8
          Removed ``dynamic`` and ``repo`` keyword arguments as they were unused.
        """
        self._ensure_unfrozen("Adding a requirement")
        self._pex_info.add_requirement(req)

    def add_from_requirements_pex(self, pex):
        """Add requirements from an existing pex.

        :param pex: The path to an existing .pex file or unzipped pex directory.
        """
        self._ensure_unfrozen("Adding from pex")
        pex_info = PexInfo.from_pex(pex)
        pex_environment = PEXEnvironment.mount(pex, pex_info=pex_info)
        for fingerprinted_dist in pex_environment.iter_distributions():
            self.add_distribution(
                dist=fingerprinted_dist.distribution, fingerprint=fingerprinted_dist.fingerprint
            )
        for requirement in pex_info.requirements:
            self.add_requirement(requirement)

    def set_executable(self, filename, env_filename=None):
        """Set the executable for this environment.

        :param filename: The file that should be executed within the PEX environment when the PEX is
          invoked.
        :keyword env_filename: (optional) The name that the executable file should be stored as within
          the PEX.  By default this will be the base name of the given filename.

        The entry point of the PEX may also be specified via ``PEXBuilder.set_entry_point``.
        """
        self._ensure_unfrozen("Setting the executable")
        if self._pex_info.script:
            raise self.InvalidExecutableSpecification(
                "Cannot set both entry point and script of PEX!"
            )
        if env_filename is None:
            env_filename = os.path.basename(filename)
        if self._chroot.get("executable"):
            raise self.InvalidExecutableSpecification(
                "Setting executable on a PEXBuilder that already has one!"
            )
        self._copy_or_link(filename, env_filename, "executable")
        entry_point = env_filename
        entry_point = entry_point.replace(os.path.sep, ".")
        self._pex_info.entry_point = entry_point.rpartition(".")[0]

    def set_script(self, script):
        """Set the entry point of this PEX environment based upon a distribution script.

        :param script: The script name as defined either by a console script or ordinary
          script within the setup.py of one of the distributions added to the PEX.
        :raises: :class:`PEXBuilder.InvalidExecutableSpecification` if the script is not found
          in any distribution added to the PEX.
        """

        distributions = OrderedSet(self._distributions.values())
        for pex in self._pex_info.pex_path:
            if os.path.exists(pex):
                distributions.update(PEX(pex, interpreter=self._interpreter).resolve())

        # Check if 'script' is a console_script.
        dist_entry_point = get_entry_point_from_console_script(script, distributions)
        if dist_entry_point:
            self.set_entry_point(str(dist_entry_point.entry_point))
            TRACER.log(
                "Set entrypoint to console_script {!r} in {!r}".format(
                    dist_entry_point.entry_point, dist_entry_point.dist
                )
            )
            return

        # Check if 'script' is an ordinary script.
        dist_script = get_script_from_distributions(script, distributions)
        if dist_script:
            if self._pex_info.entry_point:
                raise self.InvalidExecutableSpecification(
                    "Cannot set both entry point and script of PEX!"
                )
            self._pex_info.script = script
            TRACER.log("Set entrypoint to script {!r} in {!r}".format(script, dist_script.dist))
            return

        raise self.InvalidExecutableSpecification(
            "Could not find script {!r} in any distribution {} within PEX!".format(
                script, ", ".join(str(d) for d in distributions)
            )
        )

    def set_entry_point(self, entry_point):
        """Set the entry point of this PEX environment.

        :param entry_point: The entry point of the PEX in the form of ``module`` or ``module:symbol``,
          or ``None``.
        :type entry_point: string or None

        By default the entry point is None.  The behavior of a ``None`` entry point is dropping into
        an interpreter.  If ``module``, it will be executed via ``runpy.run_module``.  If
        ``module:symbol``, it is equivalent to ``from module import symbol; symbol()``.

        The entry point may also be specified via ``PEXBuilder.set_executable``.
        """
        self._ensure_unfrozen("Setting an entry point")
        self._pex_info.entry_point = entry_point

    def set_shebang(self, shebang):
        """Set the exact shebang line for the PEX file.

        For example, pex_builder.set_shebang('/home/wickman/Local/bin/python3.4').  This is
        used to override the default behavior which is to have a #!/usr/bin/env line referencing an
        interpreter compatible with the one used to build the PEX.

        :param shebang: The shebang line. If it does not include the leading '#!' it will be added.
        :type shebang: str
        """
        self._shebang = "#!%s" % shebang if not shebang.startswith("#!") else shebang

    def set_header(self, header):
        # type: (str) -> None
        """Set a header script for the PEX.

        By default, there is none and the default shebang invokes Python against the PEX zip file,
        which causes Python to look for a root `__main__.py` module in the PEX zip and execute that.
        Adding a header is not useful if the shebang selects a Python interpreter since Python will
        ignore this header script content and execute the zipapp algorithm described above. It can
        be useful though when the PEX is passed to some other type of interpreter and / or the
        shebang is also customised to be a non-Python interpreter that acts incrementally (I.E.: it
        won't try to parse the zip file all at once; thus choking on the zip content after the
        header.). An important case of this are unix shells, in particular (ba)sh, which evaluates
        files incrementally line by line.
        """
        self._header = header

    def _add_dist_dir(self, path, dist_name, fingerprint=None):
        target_dir = os.path.join(self._pex_info.internal_cache, dist_name)
        if self._copy_mode == CopyMode.SYMLINK:
            self._copy_or_link(path, target_dir, label=dist_name)
        else:
            for root, _, files in os.walk(path):
                for f in files:
                    filename = os.path.join(root, f)
                    relpath = os.path.relpath(filename, path)
                    target = os.path.join(target_dir, relpath)
                    self._copy_or_link(filename, target, label=dist_name)
        return fingerprint or CacheHelper.dir_hash(path)

    def add_distribution(
        self,
        dist,  # type: Distribution
        fingerprint=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        """Add a :class:`pex.dist_metadata.Distribution` from its handle.

        :param dist: The distribution to add to this environment.
        :keyword fingerprint: The fingerprint of the distribution, if already known.
        """
        if dist.location in self._distributions:
            TRACER.log(
                "Skipping adding {} - already added from {}".format(dist, dist.location), V=9
            )
            return
        self._ensure_unfrozen("Adding a distribution")
        dist_name = os.path.basename(dist.location)
        self._distributions[dist.location] = dist

        if not os.path.isdir(dist.location):
            raise self.InvalidDistribution(
                "Unsupported distribution type: {}, pex can only accept dist "
                "dirs (installed wheels).".format(dist)
            )
        dist_hash = self._add_dist_dir(dist.location, dist_name, fingerprint=fingerprint)

        # add dependency key so that it can rapidly be retrieved from cache
        self._pex_info.add_distribution(dist_name, dist_hash)

    def add_dist_location(
        self,
        dist,  # type: str
        fingerprint=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        """Add a distribution by its location on disk.

        :param dist: The path to the distribution to add.
        :keyword fingerprint: The fingerprint of the distribution, if already known.
        :raises PEXBuilder.InvalidDistribution: When the path does not contain a matching
          distribution.

        PEX supports only installed wheel distributions.
        """
        self._ensure_unfrozen("Adding a distribution")
        try:
            distribution = Distribution.load(dist)
        except MetadataError as e:
            raise self.InvalidDistribution(str(e))
        self.add_distribution(distribution, fingerprint=fingerprint)
        self.add_requirement(distribution.as_requirement())

    def _precompile_source(self):
        vendored_dir = os.path.join(self._pex_info.bootstrap, "pex/vendor/_vendored")
        source_relpaths = [
            path
            for label in ("source", "executable", "main", "bootstrap")
            for path in self._chroot.filesets.get(label, ())
            if path.endswith(".py")
            # N.B.: Some of our vendored code does not work with all versions of Python we support;
            # so we just skip compiling it.
            and vendored_dir != commonpath((vendored_dir, path))
        ]

        compiler = Compiler(self.interpreter)
        compiled_relpaths = compiler.compile(self._chroot.path(), source_relpaths)
        for compiled in compiled_relpaths:
            self._chroot.touch(compiled, label="bytecode")

    def _prepare_code(self):
        self._pex_info.code_hash = CacheHelper.pex_code_hash(self._chroot.path())
        self._pex_info.pex_hash = hashlib.sha1(self._pex_info.dump().encode("utf-8")).hexdigest()
        self._chroot.write(self._pex_info.dump().encode("utf-8"), PexInfo.PATH, label="manifest")

        bootstrap = BOOTSTRAP_ENVIRONMENT.format(
            bootstrap_dir=self._pex_info.bootstrap,
            pex_root=self._pex_info.raw_pex_root,
            pex_hash=self._pex_info.pex_hash,
            has_interpreter_constraints=bool(self._pex_info.interpreter_constraints),
            hermetic_venv_scripts=self._pex_info.venv_hermetic_scripts,
            pex_path=self._pex_info.pex_path,
            is_venv=self._pex_info.venv,
        )
        self._chroot.write(
            data=to_bytes(self._shebang + "\n" + self._preamble + "\n" + bootstrap),
            dst="__main__.py",
            executable=True,
            label="main",
        )
        self._chroot.write(
            data=to_bytes(bootstrap),
            dst=os.path.join("__pex__", "__init__.py"),
            label="importhook",
        )

    def _copy_or_link(self, src, dst, label=None):
        if src is None:
            self._chroot.touch(dst, label)
        elif self._copy_mode == CopyMode.COPY:
            self._chroot.copy(src, dst, label)
        elif self._copy_mode == CopyMode.SYMLINK:
            self._chroot.symlink(src, dst, label)
        else:
            self._chroot.link(src, dst, label)

    def _prepare_bootstrap(self):
        from . import vendor

        vendor.vendor_runtime(
            chroot=self._chroot,
            dest_basedir=self._pex_info.bootstrap,
            label="bootstrap",
            # NB: We use pip here in the builder, but that's only at buildtime and
            # although we don't use pyparsing directly, packaging.markers, which we
            # do use at runtime, does.
            root_module_names=["attr", "packaging", "pkg_resources", "pyparsing"],
        )
        if self._pex_info.includes_tools:
            # The `repository extract` tool needs setuptools and wheel to build sdists and wheels
            # and distutils needs .dist-info to discover setuptools (and wheel).
            vendor.vendor_runtime(
                chroot=self._chroot,
                dest_basedir=self._pex_info.bootstrap,
                label="bootstrap",
                root_module_names=["setuptools", "wheel"],
                include_dist_info=True,
            )

        bootstrap_digest = hashlib.sha1()
        bootstrap_packages = ["third_party", "venv"]
        if self._pex_info.includes_tools:
            bootstrap_packages.extend(["commands", "tools"])
        package_root = os.path.dirname(__file__)
        for root, dirs, files in os.walk(package_root):
            if root == package_root:
                dirs[:] = bootstrap_packages

            for f in filter_pyc_files(files):
                if f.endswith("testing.py"):
                    continue
                abs_src = os.path.join(root, f)
                with open(abs_src, "rb") as fp:
                    data = fp.read()
                self._chroot.write(
                    data,
                    dst=os.path.join(
                        self._pex_info.bootstrap, "pex", os.path.relpath(abs_src, package_root)
                    ),
                    label="bootstrap",
                )
                bootstrap_digest.update(data)

        self._pex_info.bootstrap_hash = bootstrap_digest.hexdigest()

    def freeze(self, bytecode_compile=True):
        """Freeze the PEX.

        :param bytecode_compile: If True, precompile .py files into .pyc files when freezing code.

        Freezing the PEX writes all the necessary metadata and environment bootstrapping code.  It may
        only be called once and renders the PEXBuilder immutable.
        """
        self._ensure_unfrozen("Freezing the environment")
        self._prepare_bootstrap()
        self._prepare_code()
        if bytecode_compile:
            self._precompile_source()
        self._frozen = True

    def build(
        self,
        path,  # type: str
        bytecode_compile=True,  # type: bool
        deterministic_timestamp=False,  # type: bool
        layout=Layout.ZIPAPP,  # type: Layout.Value
        compress=True,  # type: bool
    ):
        # type: (...) -> None
        """Package the PEX application.

        By default, the PEX is packaged as a zipapp for ease of shipping as a single file, but it
        can also be packaged in spread mode for efficiency of syncing over the network
        incrementally.

        :param path: The path where the PEX should be stored.
        :param bytecode_compile: If True, precompile .py files into .pyc files.
        :param deterministic_timestamp: If True, will use our hardcoded time for zipfile timestamps.
        :param layout: The layout to use for the PEX.
        :param compress: Whether to compress zip entries when building to a layout that uses zip
                         files.

        If the PEXBuilder is not yet frozen, it will be frozen by ``build``.  This renders the
        PEXBuilder immutable.
        """
        if not self._frozen:
            self.freeze(bytecode_compile=bytecode_compile)

        # The PEX building proceeds assuming a user will not race themselves building to a single
        # non-PEX_ROOT output path they requested;
        tmp_pex = path + "~"
        if os.path.exists(tmp_pex):
            self._logger.warning("Previous binary unexpectedly exists, cleaning: {}".format(path))
            if os.path.isfile(tmp_pex):
                os.unlink(tmp_pex)
            else:
                shutil.rmtree(tmp_pex, True)

        if layout == Layout.LOOSE:
            shutil.copytree(self.path(), tmp_pex)
        elif layout == Layout.PACKED:
            self._build_packedapp(
                dirname=tmp_pex,
                deterministic_timestamp=deterministic_timestamp,
                compress=compress,
            )
        else:
            self._build_zipapp(
                filename=tmp_pex, deterministic_timestamp=deterministic_timestamp, compress=compress
            )

        if os.path.isdir(path):
            shutil.rmtree(path, True)
        elif os.path.isdir(tmp_pex):
            safe_delete(path)
        os.rename(tmp_pex, path)

    def set_sh_boot_script(
        self,
        pex_name,  # type: str
        targets,  # type: Targets
        python_shebang,  # type: Optional[str]
    ):
        if not self._frozen:
            raise Exception("Generating a sh_boot script requires the pex to be frozen.")

        self.set_shebang("/bin/sh")
        script = create_sh_boot_script(
            pex_name=pex_name,
            pex_info=self._pex_info,
            targets=targets,
            interpreter=self.interpreter,
            python_shebang=python_shebang,
        )
        self.set_header(script)

    def _build_packedapp(
        self,
        dirname,  # type: str
        deterministic_timestamp=False,  # type: bool
        compress=True,  # type: bool
    ):
        # type: (...) -> None

        pex_info = self._pex_info.copy()
        pex_info.update(PexInfo.from_env())

        # Include user sources, PEX-INFO and __main__ as loose files in src/.
        for fileset in ("executable", "importhook", "main", "manifest", "resource", "source"):
            for f in self._chroot.filesets.get(fileset, ()):
                dest = os.path.join(dirname, f)
                safe_mkdir(os.path.dirname(dest))
                safe_copy(os.path.realpath(os.path.join(self._chroot.chroot, f)), dest)

        # Pex historically only supported compressed zips in packed layout, so we don't disturb the
        # old cache structure for those zips and instead just use a subdir for un-compressed zips.
        # This works for our two zip caches (we'll have no collisions with legacy compressed zips)
        # since the bootstrap zip has a known name that is not "un-compressed" and "un-compressed"
        # is not a valid wheel name either.
        def zip_cache_dir(path):
            # type: (str) -> str
            if compress:
                return path
            return os.path.join(path, "un-compressed")

        # Zip up the bootstrap which is constant for a given version of Pex.
        bootstrap_hash = pex_info.bootstrap_hash
        if bootstrap_hash is None:
            raise AssertionError(
                "Expected bootstrap_hash to be populated for {}.".format(self._pex_info)
            )
        cached_bootstrap_zip_dir = zip_cache_dir(
            os.path.join(pex_info.pex_root, "bootstrap_zips", bootstrap_hash)
        )
        with atomic_directory(cached_bootstrap_zip_dir) as atomic_bootstrap_zip_dir:
            if not atomic_bootstrap_zip_dir.is_finalized():
                self._chroot.zip(
                    os.path.join(atomic_bootstrap_zip_dir.work_dir, pex_info.bootstrap),
                    deterministic_timestamp=deterministic_timestamp,
                    exclude_file=is_pyc_temporary_file,
                    strip_prefix=pex_info.bootstrap,
                    labels=("bootstrap",),
                    compress=compress,
                )
        safe_copy(
            os.path.join(cached_bootstrap_zip_dir, pex_info.bootstrap),
            os.path.join(dirname, pex_info.bootstrap),
        )

        # Zip up each installed wheel chroot, which is constant for a given version of a
        # wheel.
        if pex_info.distributions:
            internal_cache = os.path.join(dirname, pex_info.internal_cache)
            os.mkdir(internal_cache)
            for location, fingerprint in pex_info.distributions.items():
                cached_installed_wheel_zip_dir = zip_cache_dir(
                    os.path.join(pex_info.pex_root, "installed_wheel_zips", fingerprint)
                )
                with atomic_directory(cached_installed_wheel_zip_dir) as atomic_zip_dir:
                    if not atomic_zip_dir.is_finalized():
                        self._chroot.zip(
                            os.path.join(atomic_zip_dir.work_dir, location),
                            deterministic_timestamp=deterministic_timestamp,
                            exclude_file=is_pyc_temporary_file,
                            strip_prefix=os.path.join(pex_info.internal_cache, location),
                            labels=(location,),
                            compress=compress,
                        )
                safe_copy(
                    os.path.join(cached_installed_wheel_zip_dir, location),
                    os.path.join(internal_cache, location),
                )

    def _build_zipapp(
        self,
        filename,  # type: str
        deterministic_timestamp=False,  # type: bool
        compress=True,  # type: bool
    ):
        # type: (...) -> None
        with safe_open(filename, "wb") as pexfile:
            assert os.path.getsize(pexfile.name) == 0
            pexfile.write(to_bytes("{}\n".format(self._shebang)))
            if self._header:
                pexfile.write(to_bytes(self._header))
        with TRACER.timed("Zipping PEX file."):
            self._chroot.zip(
                filename,
                mode="a",
                deterministic_timestamp=deterministic_timestamp,
                # When configured with a `copy_mode` of `CopyMode.SYMLINK`, we symlink distributions
                # as pointers to installed wheel directories in ~/.pex/installed_wheels/... Since
                # those installed wheels reside in a shared cache, they can be in-use by other
                # processes and so their code may be in the process of being bytecode compiled as we
                # attempt to zip up our chroot. Bytecode compilation produces ephemeral temporary
                # pyc files that we should avoid copying since they are useless and inherently
                # racy.
                exclude_file=is_pyc_temporary_file,
                compress=compress,
            )
        chmod_plus_x(filename)



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import json
import os.path
import subprocess
import sys

import pytest

from pex.cli.testing import run_pex3
from pex.interpreter import PythonInterpreter
from pex.pep_503 import ProjectName
from pex.pex import PEX
from pex.testing import (
    PY27,
    PY310,
    ensure_python_interpreter,
    ensure_python_venv,
    make_env,
    run_pex_command,
)
from pex.typing import TYPE_CHECKING
from pex.venv.virtualenv import Virtualenv
from pex.version import __version__

if TYPE_CHECKING:
    from typing import Any, Optional


def test_no_python(tmpdir):
    # type: (Any) -> None

    result = run_pex3("venv", "inspect", str(tmpdir))
    result.assert_failure()
    assert result.error.startswith(
        "The virtualenv at {venv_dir} is not valid. Failed to load an interpreter ".format(
            venv_dir=tmpdir
        )
    )


def test_not_venv_interpreter(py310):
    # type: (PythonInterpreter) -> None

    result = run_pex3("venv", "inspect", py310.binary)
    result.assert_failure()
    assert (
        "{python} is not an venv interpreter.".format(python=py310.binary) == result.error.strip()
    )


def test_not_venv_dir(py310):
    # type: (PythonInterpreter) -> None

    result = run_pex3("venv", "inspect", py310.prefix)
    result.assert_failure()
    assert "{venv} is not a venv.".format(venv=py310.prefix) == result.error.strip()


def assert_inspect(
    target,  # type: str
    expected_venv_dir,  # type: str
    expected_system_site_packages,  # type: Optional[bool]
    expected_base_interpreter,  # type: PythonInterpreter
    expected_pex_provenance,  # type: bool
):
    # type: (...) -> Any

    result = run_pex3("venv", "inspect", target)
    result.assert_success()

    data = json.loads(result.output)
    assert os.path.realpath(expected_venv_dir) == os.path.realpath(data["venv_dir"])
    assert expected_system_site_packages == data["include_system_site_packages"]

    interpreter = data["interpreter"]
    assert expected_base_interpreter.binary == interpreter["base_binary"]
    assert ".".join(map(str, expected_base_interpreter.version)) == interpreter["version"]

    provenance = data["provenance"]
    assert provenance["is_pex"] is expected_pex_provenance
    assert __version__ if expected_pex_provenance else None == provenance["pex_version"]

    if expected_pex_provenance:
        expected_created_by = (
            "virtualenv {version}".format(version=Virtualenv.VIRTUALENV_VERSION)
            if sys.version_info[0] == 2
            else "venv"
        )
        assert expected_created_by == provenance["created_by"]

    return data


def test_inspect_pex_venv(tmpdir):
    # type: (Any) -> None

    pex = os.path.join(str(tmpdir), "cowsay.pex")
    run_pex_command(
        args=["--include-tools", "cowsay==5.0", "ansicolors==1.1.8", "-o", pex]
    ).assert_success()

    venv_dir = os.path.join(str(tmpdir), "venv")
    subprocess.check_call(args=[sys.executable, pex, "venv", venv_dir], env=make_env(PEX_TOOLS=1))

    current_base_interpreter = PythonInterpreter.get().resolve_base_interpreter()
    data1 = assert_inspect(
        target=venv_dir,
        expected_venv_dir=venv_dir,
        expected_system_site_packages=False,
        expected_base_interpreter=current_base_interpreter,
        expected_pex_provenance=True,
    )
    assert "cowsay" in data1["scripts"]
    assert ["ansicolors==1.1.8", "cowsay==5.0"] == data1["distributions"]

    data2 = assert_inspect(
        target=Virtualenv(venv_dir).interpreter.binary,
        expected_venv_dir=venv_dir,
        expected_system_site_packages=False,
        expected_base_interpreter=current_base_interpreter,
        expected_pex_provenance=True,
    )
    assert data1 == data2


@pytest.mark.parametrize("python_version", [PY27, PY310])
@pytest.mark.parametrize(
    "system_site_packages",
    [pytest.param(value, id="system_site_packages={}".format(value)) for value in (True, False)],
)
def test_inspect_venv_non_pex(
    python_version,  # type: str
    system_site_packages,  # type: bool
):
    # type: (...) -> None

    python, _ = ensure_python_venv(python_version, system_site_packages=system_site_packages)
    python_major_version = int(python_version.split(".")[0])
    assert_inspect(
        target=python,
        expected_venv_dir=os.path.dirname(os.path.dirname(python)),
        expected_system_site_packages=None if python_major_version == 2 else system_site_packages,
        expected_base_interpreter=PythonInterpreter.from_binary(
            ensure_python_interpreter(python_version)
        ),
        expected_pex_provenance=False,
    )


@pytest.mark.parametrize(
    "system_site_packages",
    [pytest.param(value, id="system_site_packages={}".format(value)) for value in (True, False)],
)
def test_inspect_venv_virtualenv(
    tmpdir,  # type: Any
    system_site_packages,  # type: bool
):
    # type: (...) -> None

    pex = os.path.join(str(tmpdir), "virtualenv.pex")
    run_pex_command(args=["virtualenv", "-c", "virtualenv", "-o", pex]).assert_success()
    dists = {dist.metadata.project_name: dist.metadata.version for dist in PEX(pex).resolve()}
    virtualenv_version = dists[ProjectName("virtualenv")]

    venv_dir = os.path.join(str(tmpdir), "venv")
    args = [sys.executable, pex, venv_dir]
    if system_site_packages:
        args.append("--system-site-packages")
    subprocess.check_call(args=args)

    data = assert_inspect(
        target=venv_dir,
        expected_venv_dir=venv_dir,
        expected_system_site_packages=system_site_packages,
        expected_base_interpreter=PythonInterpreter.get().resolve_base_interpreter(),
        expected_pex_provenance=False,
    )
    assert "pip" in data["scripts"]

    provenance = data["provenance"]
    assert "virtualenv {version}".format(version=virtualenv_version.raw) == provenance["created_by"]



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os.path
from textwrap import dedent

import pytest

from pex.common import touch
from pex.interpreter import PyVenvCfg
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_parse_invalid(tmpdir):
    # type: (Any) -> None

    pyvenv_cfg = os.path.join(str(tmpdir), "pyvenv.cfg")
    touch(pyvenv_cfg)
    with pytest.raises(PyVenvCfg.Error):
        PyVenvCfg.parse(pyvenv_cfg)


def test_parse_nominal(tmpdir):
    # type: (Any) -> None

    pyvenv_cfg = os.path.join(str(tmpdir), "pyvenv.cfg")
    with open(pyvenv_cfg, "w") as fp:
        fp.write(
            dedent(
                """\
                home = foo
                include-system-site-packages = true
                version = 1.2.3
                """
            )
        )

    cfg = PyVenvCfg.parse(pyvenv_cfg)
    assert pyvenv_cfg == cfg.path
    assert "foo" == cfg.home
    assert cfg.include_system_site_packages is True
    assert "1.2.3" == cfg.config("version")
    assert cfg.config("bar") is None
    assert "baz" == cfg.config("bar", "baz")


def test_find_not_found(tmpdir):
    # type: (Any) -> None

    python = os.path.join(str(tmpdir), "bin", "python")
    touch(python)
    assert PyVenvCfg.find(python) is None


def test_find_python_sibling(tmpdir):
    # type: (Any) -> None

    bin_dir = os.path.join(str(tmpdir), "bin")
    python = os.path.join(bin_dir, "python")
    touch(python)
    with open(os.path.join(bin_dir, "pyvenv.cfg"), "w") as fp:
        fp.write("home = foo")
    cfg = PyVenvCfg.find(python)
    assert cfg is not None
    assert "foo" == cfg.home


def test_find_venv_sibling(tmpdir):
    # type: (Any) -> None

    python = os.path.join(str(tmpdir), "bin", "python")
    touch(python)
    with open(os.path.join(str(tmpdir), "pyvenv.cfg"), "w") as fp:
        fp.write("home = foo")
    cfg = PyVenvCfg.find(python)
    assert cfg is not None
    assert "foo" == cfg.home



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import logging
from argparse import ArgumentParser, _ActionsContainer

from pex.cli.command import BuildTimeCommand
from pex.commands.command import JsonMixin, OutputMixin
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraint
from pex.resolve import target_options
from pex.resolve.target_configuration import InterpreterConstraintsNotSatisfied, InterpreterNotFound
from pex.result import Error, Ok, Result
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Dict


logger = logging.getLogger(__name__)


class Interpreter(OutputMixin, JsonMixin, BuildTimeCommand):
    @classmethod
    def _add_inspect_arguments(cls, parser):
        # type: (_ActionsContainer) -> None
        cls.add_output_option(parser, entity="Python interpreter path")
        parser.add_argument(
            "-v",
            "--verbose",
            action="store_true",
            help=(
                "Provide more information about the interpreter in JSON format. Use --tags and "
                "--markers to include even more information. The 'path' key is guaranteed to be "
                "present at a minimum and will contain the path to the interpreter."
            ),
        )
        parser.add_argument(
            "-t",
            "--tags",
            "--compatible-tags",
            action="store_true",
            help=(
                "Include the interpreter's compatible tags in the verbose JSON output under the "
                "'compatible_tags' key with more preferable (more specific) tags ordered earlier. "
                "The 'path' key is guaranteed to be present and will contain the path to the "
                "interpreter."
            ),
        )
        parser.add_argument(
            "-m",
            "--markers",
            "--marker-env",
            "--marker-environment",
            action="store_true",
            help=(
                "Include the interpreter's PEP-508 marker environment in the verbose JSON output "
                "under the 'marker_environment' key. The 'path' key is guaranteed to be present "
                "and will contain the path to the interpreter."
            ),
        )
        cls.add_json_options(parser, entity="verbose output")

        interpreter_options_parser = parser.add_argument_group(
            title="Interpreter options",
            description=(
                "Specify which interpreters to inspect. The current interpreter is inspected "
                "by default."
            ),
        )
        interpreter_options_parser.add_argument(
            "-a",
            "--all",
            action="store_true",
            help="Print all compatible interpreters, preferred first.",
        )
        target_options.register(interpreter_options_parser, include_platforms=False)

    @classmethod
    def add_extra_arguments(
        cls,
        parser,  # type: ArgumentParser
    ):
        # type: (...) -> None
        subcommands = cls.create_subcommands(
            parser,
            description="Interact with local interpreters via the following subcommands.",
        )
        with subcommands.parser(
            name="inspect",
            help="Inspect local interpreters",
            func=cls._inspect,
            include_verbosity=False,
        ) as inspect_parser:
            cls._add_inspect_arguments(inspect_parser)

    def _inspect(self):
        # type: () -> Result

        interpreter_configuration = target_options.configure_interpreters(self.options)
        try:
            interpreters = interpreter_configuration.resolve_interpreters()
        except (InterpreterNotFound, InterpreterConstraintsNotSatisfied) as e:
            return Error(str(e))

        if self.options.all:
            interpreters.update(PythonInterpreter.all(paths=interpreter_configuration.python_path))
        if not interpreters:
            interpreters.add(PythonInterpreter.get())

        verbose = self.options.verbose or self.options.tags or self.options.markers
        if self.options.indent and not verbose:
            logger.warning(
                "Ignoring --indent={} since --verbose mode is not enabled.".format(
                    self.options.indent
                )
            )
        with self.output(self.options) as out:
            for interpreter in interpreters:
                if verbose:
                    interpreter_info = {"path": interpreter.binary}  # type: Dict[str, Any]
                    if self.options.verbose:
                        interpreter_info.update(
                            version=interpreter.identity.version_str,
                            requirement=str(InterpreterConstraint.exact_version(interpreter)),
                            platform=str(interpreter.platform),
                            venv=interpreter.is_venv,
                        )
                        if interpreter.is_venv:
                            try:
                                interpreter_info[
                                    "base_interpreter"
                                ] = interpreter.resolve_base_interpreter().binary
                            except PythonInterpreter.BaseInterpreterResolutionError as e:
                                logger.warning(
                                    "Failed to determine base interpreter for venv interpreter "
                                    "{interpreter}: {err}".format(
                                        interpreter=interpreter.binary, err=e
                                    )
                                )
                            interpreter_info[
                                "base_interpreter"
                            ] = interpreter.resolve_base_interpreter().binary
                    if self.options.tags:
                        interpreter_info[
                            "compatible_tags"
                        ] = interpreter.identity.supported_tags.to_string_list()
                    if self.options.markers:
                        interpreter_info[
                            "marker_environment"
                        ] = interpreter.identity.env_markers.as_dict()
                    self.dump_json(self.options, interpreter_info, out)
                else:
                    out.write(interpreter.binary)
                out.write("\n")

        return Ok()



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import filecmp
import os
import stat
import subprocess
import sys
import zipfile

import pytest

from pex.common import open_zip, safe_open, temporary_dir, touch
from pex.compatibility import WINDOWS, commonpath
from pex.executor import Executor
from pex.layout import Layout
from pex.pex import PEX
from pex.pex_builder import CopyMode, PEXBuilder
from pex.testing import WheelBuilder, install_wheel, make_bdist, make_env
from pex.testing import write_simple_pex as write_pex
from pex.typing import TYPE_CHECKING
from pex.variables import ENV

if TYPE_CHECKING:
    from typing import Any, Iterator, List, Set

exe_main = """
import sys
from p1.my_module import do_something
do_something()

with open(sys.argv[1], 'w') as fp:
  fp.write('success')
"""


def test_pex_builder():
    # type: () -> None
    with temporary_dir() as td, make_bdist("p1") as p1:
        pb = write_pex(td, exe_main, dists=[p1])

        success_txt = os.path.join(td, "success.txt")
        PEX(td, interpreter=pb.interpreter).run(args=[success_txt])
        assert os.path.exists(success_txt)
        with open(success_txt) as fp:
            assert fp.read() == "success"


def test_pex_builder_shebang():
    # type: () -> None
    def builder(shebang):
        # type: (str) -> PEXBuilder
        pb = PEXBuilder()
        pb.set_shebang(shebang)
        return pb

    for pb in builder("foobar"), builder("#!foobar"):
        for b in pb, pb.clone():
            with temporary_dir() as td:
                target = os.path.join(td, "foo.pex")
                b.build(target)
                expected_preamble = b"#!foobar\n"
                with open(target, "rb") as fp:
                    assert fp.read(len(expected_preamble)) == expected_preamble


def test_pex_builder_preamble():
    # type: () -> None
    with temporary_dir() as td:
        target = os.path.join(td, "foo.pex")
        should_create = os.path.join(td, "foo.1")

        tempfile_preamble = "\n".join(
            ["import sys", "open('{0}', 'w').close()".format(should_create), "sys.exit(3)"]
        )

        pb = PEXBuilder(preamble=tempfile_preamble)
        pb.build(target)

        assert not os.path.exists(should_create)

        pex = PEX(target, interpreter=pb.interpreter)
        process = pex.run(blocking=False)
        process.wait()

        assert process.returncode == 3
        assert os.path.exists(should_create)


def test_pex_builder_compilation():
    # type: () -> None
    with temporary_dir() as td1, temporary_dir() as td2, temporary_dir() as td3:
        src = os.path.join(td1, "src.py")
        with open(src, "w") as fp:
            fp.write(exe_main)

        exe = os.path.join(td1, "exe.py")
        with open(exe, "w") as fp:
            fp.write(exe_main)

        def build_and_check(path, precompile):
            # type: (str, bool) -> None
            pb = PEXBuilder(path=path)
            pb.add_source(src, "lib/src.py")
            pb.set_executable(exe, "exe.py")
            pb.freeze(bytecode_compile=precompile)
            for pyc_file in ("exe.pyc", "lib/src.pyc", "__main__.pyc"):
                pyc_exists = os.path.exists(os.path.join(path, pyc_file))
                if precompile:
                    assert pyc_exists
                else:
                    assert not pyc_exists
            bootstrap_dir = os.path.join(path, pb.info.bootstrap)
            bootstrap_pycs = []  # type: List[str]
            for _, _, files in os.walk(bootstrap_dir):
                bootstrap_pycs.extend(f for f in files if f.endswith(".pyc"))
            if precompile:
                assert len(bootstrap_pycs) > 0
            else:
                assert 0 == len(bootstrap_pycs)

        build_and_check(td2, False)
        build_and_check(td3, True)


@pytest.mark.skipif(WINDOWS, reason="No hardlinks on windows")
def test_pex_builder_copy_or_link():
    # type: () -> None
    with temporary_dir() as td:
        src = os.path.join(td, "exe.py")
        with safe_open(src, "w") as fp:
            fp.write(exe_main)

        def build_and_check(copy_mode):
            # type: (CopyMode.Value) -> None
            pb = PEXBuilder(copy_mode=copy_mode)
            path = pb.path()
            pb.add_source(src, "exe.py")

            path_clone = os.path.join(path, "__clone")
            pb.clone(into=path_clone)

            for root in path, path_clone:
                s1 = os.stat(src)
                s2 = os.stat(os.path.join(root, "exe.py"))
                is_link = (s1[stat.ST_INO], s1[stat.ST_DEV]) == (s2[stat.ST_INO], s2[stat.ST_DEV])
                if copy_mode == CopyMode.COPY:
                    assert not is_link
                else:
                    # Since os.stat follows symlinks; so in CopyMode.SYMLINK, this just proves
                    # the symlink points to the original file. Going further and checking path
                    # and path_clone for the presence of a symlink (an os.islink test) is
                    # trickier since a Linux hardlink of a symlink produces a symlink whereas a
                    # macOS hardlink of a symlink produces a hardlink.
                    assert is_link

        build_and_check(CopyMode.LINK)
        build_and_check(CopyMode.COPY)
        build_and_check(CopyMode.SYMLINK)


@pytest.fixture
def tmp_chroot(tmpdir):
    # type: (Any) -> Iterator[str]
    tmp_chroot = str(tmpdir)
    cwd = os.getcwd()
    try:
        os.chdir(tmp_chroot)
        yield tmp_chroot
    finally:
        os.chdir(cwd)


@pytest.mark.parametrize(
    "copy_mode", [pytest.param(copy_mode, id=copy_mode.value) for copy_mode in CopyMode.values()]
)
def test_pex_builder_add_source_relpath_issues_1192(
    tmp_chroot,  # type: str
    copy_mode,  # type: CopyMode.Value
):
    # type: (...) -> None
    pb = PEXBuilder(copy_mode=copy_mode)
    with safe_open("src/main.py", "w") as fp:
        fp.write("import sys; sys.exit(42)")
    pb.add_source("src/main.py", "main.py")
    pb.set_entry_point("main")
    pb.build("test.pex")

    process = Executor.open_process(cmd=[os.path.abspath("test.pex")])
    process.wait()
    assert 42 == process.returncode


def test_pex_builder_deterministic_timestamp():
    # type: () -> None
    pb = PEXBuilder()
    with temporary_dir() as td:
        target = os.path.join(td, "foo.pex")
        pb.build(target, deterministic_timestamp=True)
        with zipfile.ZipFile(target) as zf:
            assert all(zinfo.date_time == (1980, 1, 1, 0, 0, 0) for zinfo in zf.infolist())


def test_pex_builder_from_requirements_pex():
    # type: () -> None
    def build_from_req_pex(path, req_pex):
        # type: (str, str) -> PEXBuilder
        pb = PEXBuilder(path=path)
        pb.add_from_requirements_pex(req_pex)
        with open(os.path.join(path, "exe.py"), "w") as fp:
            fp.write(exe_main)
        pb.set_executable(os.path.join(path, "exe.py"))
        pb.freeze()
        return pb

    def verify(pb):
        # type: (PEXBuilder) -> None
        success_txt = os.path.join(pb.path(), "success.txt")
        PEX(pb.path(), interpreter=pb.interpreter).run(args=[success_txt])
        assert os.path.exists(success_txt)
        with open(success_txt) as fp:
            assert fp.read() == "success"

    # Build from pex dir.
    with temporary_dir() as td2:
        with temporary_dir() as td1, make_bdist("p1") as p1:
            pb1 = write_pex(td1, dists=[p1])
            pb2 = build_from_req_pex(td2, pb1.path())
        verify(pb2)

    # Build from .pex file.
    with temporary_dir() as td4:
        with temporary_dir() as td3, make_bdist("p1") as p1:
            pb3 = write_pex(td3, dists=[p1])
            target = os.path.join(td3, "foo.pex")
            pb3.build(target)
            pb4 = build_from_req_pex(td4, target)
        verify(pb4)


def test_pex_builder_script_from_pex_path(tmpdir):
    # type: (Any) -> None

    pex_with_script = os.path.join(str(tmpdir), "script.pex")
    with make_bdist(
        name="my_project",
        entry_points={"console_scripts": ["my_app = my_project.my_module:do_something"]},
    ) as dist:
        pb = PEXBuilder()
        pb.add_dist_location(dist.location)
        pb.build(pex_with_script)

    pex_file = os.path.join(str(tmpdir), "app.pex")
    pb = PEXBuilder()
    pb.info.pex_path = [pex_with_script]
    pb.set_script("my_app")
    pb.build(pex_file)

    assert "hello world!\n" == subprocess.check_output(args=[pex_file]).decode("utf-8")


def test_pex_builder_setuptools_script(tmpdir):
    # type: (Any) -> None

    pex_file = os.path.join(str(tmpdir), "app.pex")
    with make_bdist(
        name="my_project",
    ) as dist:
        pb = PEXBuilder()
        pb.add_dist_location(dist.location)
        pb.set_script("shell_script")
        pb.build(pex_file)

    assert "hello world from shell script\n" == subprocess.check_output(args=[pex_file]).decode(
        "utf-8"
    )


def test_pex_builder_packed(tmpdir):
    # type: (Any) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")
    pex_app = os.path.join(str(tmpdir), "app.pex")
    source_file = os.path.join(str(tmpdir), "src")
    touch(source_file)

    with ENV.patch(PEX_ROOT=pex_root), make_bdist(name="my_project") as dist:
        pb = PEXBuilder(copy_mode=CopyMode.SYMLINK)
        pb.add_source(source_file, "a.file")
        pb.add_dist_location(dist.location)
        pb.set_script("shell_script")
        pb.build(pex_app, layout=Layout.PACKED)

    assert "hello world from shell script\n" == subprocess.check_output(
        args=[os.path.join(pex_app, "__main__.py")]
    ).decode("utf-8")

    spread_dist_bootstrap = os.path.join(pex_app, pb.info.bootstrap)
    assert zipfile.is_zipfile(spread_dist_bootstrap)

    cached_bootstrap_zip = os.path.join(
        pex_root, "bootstrap_zips", pb.info.bootstrap_hash, pb.info.bootstrap
    )
    assert zipfile.is_zipfile(cached_bootstrap_zip)

    assert filecmp.cmp(spread_dist_bootstrap, cached_bootstrap_zip, shallow=False)

    assert os.path.isfile(os.path.join(pex_app, "a.file"))
    for root, dirs, files in os.walk(pex_app, followlinks=False):
        for f in files:
            path = os.path.join(root, f)
            assert not os.path.islink(path) or pex_app == commonpath(
                [pex_app, os.path.realpath(path)]
            ), (
                "All packed layout files should be real files inside the packed layout root that "
                "are divorced from either the PEXBuilder chroot or PEX_ROOT caches."
            )

    assert 1 == len(pb.info.distributions)
    location, sha = next(iter(pb.info.distributions.items()))

    spread_dist_zip = os.path.join(pex_app, pb.info.internal_cache, location)
    assert zipfile.is_zipfile(spread_dist_zip)

    cached_dist_zip = os.path.join(pex_root, "installed_wheel_zips", sha, location)
    assert zipfile.is_zipfile(cached_dist_zip)

    assert filecmp.cmp(spread_dist_zip, cached_dist_zip, shallow=False)


@pytest.mark.parametrize(
    "copy_mode", [pytest.param(copy_mode, id=copy_mode.value) for copy_mode in CopyMode.values()]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
def test_pex_builder_exclude_bootstrap_testing(
    tmpdir,  # type: Any
    copy_mode,  # type: CopyMode.Value
    layout,  # type: Layout.Value
):
    # type: (...) -> None

    pex_path = os.path.join(str(tmpdir), "empty.pex")
    pb = PEXBuilder(copy_mode=copy_mode)
    pb.build(pex_path, layout=layout)

    bootstrap_location = os.path.join(pex_path, pb.info.bootstrap)
    bootstrap_files = set()  # type: Set[str]
    if Layout.ZIPAPP == layout:
        with open_zip(pex_path) as zf:
            bootstrap_files.update(
                os.path.relpath(f, pb.info.bootstrap)
                for f in zf.namelist()
                if f.startswith(pb.info.bootstrap)
            )
    elif Layout.PACKED == layout:
        with open_zip(bootstrap_location) as zf:
            bootstrap_files.update(zf.namelist())
    else:
        bootstrap_files.update(
            os.path.relpath(os.path.join(root, f), bootstrap_location)
            for root, _, files in os.walk(bootstrap_location)
            for f in files
        )

    assert {"pex/pex_bootstrapper.py", "pex/pex_info.py", "pex/pex.py"}.issubset(
        bootstrap_files
    ), "Expected the `.bootstrap` to contain at least some of the key Pex runtime modules."
    assert not [
        f for f in bootstrap_files if f.endswith(("testing.py", "testing.pyc"))
    ], "Expected testing support files to be stripped from the Pex `.bootstrap`."


@pytest.mark.parametrize(
    "strip_pex_env", [pytest.param(True, id="StripPexEnv"), pytest.param(False, id="NoStripPexEnv")]
)
@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in Layout.values()]
)
@pytest.mark.parametrize("venv", [pytest.param(True, id="VENV"), pytest.param(False, id="UNZIP")])
def test_pex_env_var_issues_1485(
    tmpdir,  # type: Any
    strip_pex_env,  # type: bool
    venv,  # type: bool
    layout,  # type: Layout.Value
):
    # type: (...) -> None

    pex_path = os.path.join(str(tmpdir), "empty.pex")
    pb = PEXBuilder()
    pb.info.strip_pex_env = strip_pex_env
    pb.info.venv = venv
    pb.build(pex_path, layout=layout)

    launch_args = [pex_path] if layout == Layout.ZIPAPP else [sys.executable, pex_path]
    pex_root = os.path.join(str(tmpdir), "pex_root")

    def assert_pex_env_var(
        script,  # type: str
        expected_pex_env_var,  # type: str
        expect_empty_pex_root,  # type: bool
        expected_layout=layout,  # type: Layout.Value
    ):
        # type: (...) -> None
        if expect_empty_pex_root:
            assert not os.path.exists(pex_root)
        else:
            assert len(os.listdir(pex_root)) > 0
        output = subprocess.check_output(
            launch_args + ["-c", script], env=make_env(PEX_ROOT=pex_root)
        )
        actual = output.decode("utf-8").strip()
        assert os.path.realpath(expected_pex_env_var) == actual
        if Layout.ZIPAPP == expected_layout:
            assert zipfile.is_zipfile(actual)
        else:
            assert os.path.isdir(actual)

    print_pex_env_var_script = "import os; print(os.environ['PEX'])"

    assert_pex_env_var(
        script=print_pex_env_var_script,
        expected_pex_env_var=pex_path,
        expect_empty_pex_root=True,
    )
    assert_pex_env_var(
        script=print_pex_env_var_script,
        expected_pex_env_var=pex_path,
        expect_empty_pex_root=False,
    )

    other_pex_path = os.path.join(str(tmpdir), "other.pex")
    other_pb = PEXBuilder()
    other_pb.info.includes_tools = True
    other_pex_main = os.path.join(str(tmpdir), "main.py")
    with open(other_pex_main, mode="w") as fp:
        fp.write(print_pex_env_var_script)
    other_pb.set_executable(other_pex_main)
    other_pb.build(other_pex_path, layout=Layout.ZIPAPP)

    def assert_pex_env_var_nested(**env):
        # type: (**Any) -> None
        assert_pex_env_var(
            script="import subprocess; subprocess.check_call([{other_pex!r}], env={env!r})".format(
                other_pex=other_pex_path, env=make_env(**env)
            ),
            expected_pex_env_var=other_pex_path,
            expect_empty_pex_root=False,
            expected_layout=Layout.ZIPAPP,
        )

    assert_pex_env_var_nested()
    assert_pex_env_var_nested(PEX_VENV=False)
    assert_pex_env_var_nested(PEX_VENV=True)


@pytest.mark.parametrize(
    "layout", [pytest.param(layout, id=layout.value) for layout in (Layout.PACKED, Layout.ZIPAPP)]
)
def test_build_compression(
    tmpdir,  # type: Any
    layout,  # type: Layout.Value
    pex_project_dir,  # type: str
):
    # type: (...) -> None

    pex_root = os.path.join(str(tmpdir), "pex_root")

    pb = PEXBuilder()
    pb.info.pex_root = pex_root
    with ENV.patch(PEX_ROOT=pex_root):
        pb.add_dist_location(install_wheel(WheelBuilder(pex_project_dir).bdist()).location)
    exe = os.path.join(str(tmpdir), "exe.py")
    with open(exe, "w") as fp:
        fp.write("import pex; print(pex.__file__)")
    pb.set_executable(exe)

    def assert_pex(pex):
        # type: (str) -> None
        assert (
            subprocess.check_output(args=[sys.executable, pex]).decode("utf-8").startswith(pex_root)
        )

    compressed_pex = os.path.join(str(tmpdir), "compressed.pex")
    pb.build(compressed_pex, layout=layout)
    assert_pex(compressed_pex)

    uncompressed_pex = os.path.join(str(tmpdir), "uncompressed.pex")
    pb.build(uncompressed_pex, layout=layout, compress=False)
    assert_pex(uncompressed_pex)

    def size(pex):
        # type: (str) -> int
        if os.path.isfile(pex):
            return os.path.getsize(pex)
        return sum(
            os.path.getsize(os.path.join(root, f)) for root, _, files in os.walk(pex) for f in files
        )

    assert size(compressed_pex) < size(uncompressed_pex)



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from argparse import ArgumentError, ArgumentTypeError

from pex.cli import commands
from pex.cli.command import BuildTimeCommand
from pex.commands.command import GlobalConfigurationError, Main
from pex.result import catch
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union


class Pex3(Main[BuildTimeCommand]):
    """Tools for building and working with [P]ython [EX]ecutables.

    N.B: This interface is considered unstable until the release of Pex 3.0 at which point it will
    replace the current `pex` command.
    """


def main():
    # type: () -> Union[int, str]

    pex3 = Pex3(command_types=commands.all_commands())
    try:
        with pex3.parsed_command() as command:
            result = catch(command.run)
            result.maybe_display()
            return result.exit_code
    except (ArgumentError, ArgumentTypeError, GlobalConfigurationError) as e:
        return str(e)



# Copyright 2023 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).
import json
import os
import re
import subprocess
import sys
import tempfile
from textwrap import dedent

import pytest

from pex.build_system import pep_517
from pex.cli.testing import run_pex3
from pex.dist_metadata import Requirement
from pex.pep_440 import Version
from pex.pep_503 import ProjectName
from pex.pip.version import PipVersion
from pex.resolve.configured_resolver import ConfiguredResolver
from pex.resolve.locked_resolve import LockedRequirement
from pex.resolve.lockfile import json_codec
from pex.resolve.path_mappings import PathMapping, PathMappings
from pex.result import try_
from pex.sorted_tuple import SortedTuple
from pex.targets import LocalInterpreter
from pex.testing import (
    IS_LINUX,
    PY310,
    PY_VER,
    ensure_python_interpreter,
    make_env,
    run_pex_command,
)
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any, Callable, Dict

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class Repo(object):
    find_links = attr.ib()  # type: str
    path_mapping = attr.ib()  # type: PathMapping
    path_mappings = attr.ib(init=False)  # type: PathMappings
    path_mapping_arg = attr.ib(init=False)  # type: str

    @path_mappings.default
    def _mappings(self):
        # type: () -> PathMappings
        return PathMappings((self.path_mapping,))

    @path_mapping_arg.default
    def _mapping_arg(self):
        # type: () -> str
        return "{name}|{path}".format(name=self.path_mapping.name, path=self.path_mapping.path)


@pytest.fixture
def build_sdist(tmpdir):
    # type: (Any) -> Callable[[str], Repo]

    def func(project_directory):
        find_links = os.path.join(str(tmpdir), "find-links")
        os.makedirs(find_links)
        try_(
            pep_517.build_sdist(
                project_directory=project_directory,
                dist_dir=find_links,
                pip_version=PipVersion.VENDORED,
                target=LocalInterpreter.create(),
                resolver=ConfiguredResolver.default(),
            )
        )
        return Repo(find_links=find_links, path_mapping=PathMapping(path=find_links, name="FL"))

    return func


@pytest.mark.skipif(
    sys.version_info[0] < 3,
    reason="Encoding of setup.py files for Python 2 is tricky and not worth the trouble.",
)
def test_lock_uncompilable_sdist(
    tmpdir,  # type: Any
    build_sdist,  # type: Callable[[str], Repo]
):
    # type: (...) -> None

    project = os.path.join(str(tmpdir), "project")
    os.mkdir(project)
    with open(os.path.join(project, "bad.c"), "w") as fp:
        fp.write("This is not valid C code.")

    with open(os.path.join(project, "README"), "w") as fp:
        fp.write("This is a Python C-extension project that does not compile.")

    with open(os.path.join(project, "setup.py"), "w") as fp:
        fp.write(
            dedent(
                """\
                import json
                import os

                from setuptools import setup, Extension


                setup_kwargs = dict(
                    name="pex.tests.bad-c-extension",
                    version="0.1.0+test",
                    author="John Sirois",
                    author_email="js@example.com",
                    url="http://example.com/bad",
                    ext_modules=[Extension("bad", sources=["bad.c"])],
                )
                setup_kwargs.update(json.loads(os.environ.get("SETUP_KWARGS_JSON", "{}")))
                setup(**setup_kwargs)
                """
            )
        )

    repo = build_sdist(project)

    lock = os.path.join(str(tmpdir), "lock.json")
    run_pex3(
        "lock",
        "create",
        "-f",
        repo.find_links,
        "pex.tests.bad-c-extension",
        "--path-mapping",
        repo.path_mapping_arg,
        "--indent",
        "2",
        "-o",
        lock,
        env=make_env(
            SETUP_KWARGS_JSON=json.dumps(
                dict(install_requires=["ansicolors==1.1.8"], python_requires=">=3.5,<3.12")
            )
        ),
    ).assert_success()

    lockfile = json_codec.load(lockfile_path=lock, path_mappings=repo.path_mappings)
    assert len(lockfile.locked_resolves) == 1
    locked_resolve = lockfile.locked_resolves[0]
    locked_requirements = {
        locked_requirement.pin.project_name: locked_requirement
        for locked_requirement in locked_resolve.locked_requirements
    }  # type: Dict[ProjectName, LockedRequirement]
    bad = locked_requirements.pop(ProjectName("pex.tests.bad-c-extension"))
    assert Version("0.1.0+test") == bad.pin.version
    assert SpecifierSet(">=3.5,<3.12") == bad.requires_python
    assert SortedTuple([Requirement.parse("ansicolors==1.1.8")]) == bad.requires_dists
    assert locked_requirements.pop(ProjectName("ansicolors")) is not None
    assert not locked_requirements

    result = run_pex_command(args=["--lock", lock, "--path-mapping", repo.path_mapping_arg])
    result.assert_failure()
    assert "pex.tests.bad-c-extension-0.1.0+test.tar.gz" in result.error, result.error
    assert "ERROR: Failed to build one or more wheels" in result.error, result.error


@pytest.mark.skipif(
    not IS_LINUX or PY_VER < (3, 7),
    reason=(
        "The evdev project requires Linux and Python 3 and we use a setuptools in our in-tree "
        "build backend that requires Python 3.7+."
    ),
)
def test_pep_517_prepare_metadata_for_build_wheel_fallback(
    tmpdir,  # type: Any
    build_sdist,  # type: Callable[[str], Repo]
):
    # type: (...) -> None

    python = ensure_python_interpreter(PY310)

    evdev = os.path.join(str(tmpdir), "python-evdev")
    os.mkdir(evdev)
    subprocess.check_call(args=["git", "init"], cwd=evdev)
    evdev_1_6_1_sha = "2dd6ce6364bb67eedb209f6aa0bace0c18a3a40a"
    subprocess.check_call(
        args=[
            "git",
            "fetch",
            "--depth",
            "1",
            "https://github.com/gvalkov/python-evdev",
            evdev_1_6_1_sha,
        ],
        cwd=evdev,
    )
    subprocess.check_call(args=["git", "reset", "--hard", evdev_1_6_1_sha], cwd=evdev)
    with tempfile.NamedTemporaryFile() as fp:
        fp.write(
            dedent(
                """\
                diff --git a/builder/delegate_to_setuptools.py b/builder/delegate_to_setuptools.py
                new file mode 100644
                index 0000000..9a4d93d
                --- /dev/null
                +++ b/builder/delegate_to_setuptools.py
                @@ -0,0 +1,6 @@
                +from setuptools import build_meta
                +
                +
                +build_sdist = build_meta.build_sdist
                +build_wheel = build_meta.build_wheel
                +
                diff --git a/pyproject.toml b/pyproject.toml
                new file mode 100644
                index 0000000..7c52595
                --- /dev/null
                +++ b/pyproject.toml
                @@ -0,0 +1,5 @@
                +[build-system]
                +requires = ["setuptools==67.2.0", "wheel==0.38.4"]
                +backend-path = ["builder"]
                +build-backend = "delegate_to_setuptools"
                +
                diff --git a/setup.py b/setup.py
                index 73ba1f5..c19fa76 100755
                --- a/setup.py
                +++ b/setup.py
                @@ -41,7 +41,7 @@ ecodes_c = Extension('evdev._ecodes', sources=['evdev/ecodes.c'], extra_compile_
                 #-----------------------------------------------------------------------------
                 kw = {
                     'name':                 'evdev',
                -    'version':              '1.6.1',
                +    'version':              '1.6.1+test',

                     'description':          'Bindings to the Linux input handling subsystem',
                     'long_description':     (curdir / 'README.rst').read_text(),
                @@ -53,7 +53,7 @@ kw = {
                     'url':                  'https://github.com/gvalkov/python-evdev',
                     'classifiers':          classifiers,

                -    'packages':             ['evdev'],
                +    'packages':             ['evdev', 'builder'],
                     'ext_modules':          [input_c, uinput_c, ecodes_c],
                     'include_package_data': False,
                     'zip_safe':             True,
                """
            ).encode("utf-8")
        )
        fp.flush()
        subprocess.check_call(args=["git", "apply", fp.name], cwd=evdev)

    repo = build_sdist(evdev)

    lock = os.path.join(str(tmpdir), "lock.json")
    result = run_pex3(
        "lock",
        "create",
        "-vvv",
        "--python",
        python,
        "-f",
        repo.find_links,
        "evdev==1.6.1+test",
        "--path-mapping",
        repo.path_mapping_arg,
        "--indent",
        "2",
        "-o",
        lock,
    )
    result.assert_success()

    assert re.search(
        r"Failed to prepare metadata for .+{}, trying to build a wheel instead: ".format(
            re.escape("evdev-1.6.1+test.tar.gz")
        ),
        result.error,
    ), result.error



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import re

import pytest

from pex import targets
from pex.dist_metadata import Requirement
from pex.interpreter import PythonInterpreter
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.targets import (
    AbbreviatedPlatform,
    CompletePlatform,
    LocalInterpreter,
    RequiresPythonError,
    Target,
    Targets,
)
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Optional


def test_current(current_interpreter):
    # type: (PythonInterpreter) -> None
    assert LocalInterpreter.create() == targets.current()
    assert LocalInterpreter.create(current_interpreter) == targets.current()


def test_interpreter(
    py27,  # type: PythonInterpreter
    current_interpreter,  # type: PythonInterpreter
):
    # type: (...) -> None
    assert Targets().interpreter is None
    assert py27 == Targets(interpreters=(py27,)).interpreter
    assert py27 == Targets(interpreters=(py27, current_interpreter)).interpreter


def test_unique_targets(
    py27,  # type: PythonInterpreter
    py38,  # type: PythonInterpreter
    py310,  # type: PythonInterpreter
    current_interpreter,  # type: PythonInterpreter
    current_platform,  # type: Platform
):
    # type: (...) -> None
    assert (
        OrderedSet([targets.current()]) == Targets().unique_targets()
    ), "Expected the default TargetConfiguration to produce the current interpreter."

    assert OrderedSet([targets.current()]) == Targets(platforms=(None,)).unique_targets(), (
        "Expected the 'current' platform - which maps to `None` - to produce the current "
        "interpreter when no interpreters were configured."
    )

    assert (
        OrderedSet([LocalInterpreter.create(py27)])
        == Targets(interpreters=(py27,), platforms=(None,)).unique_targets()
    ), (
        "Expected the 'current' platform - which maps to `None` - to be ignored when at least one "
        "concrete interpreter for the current platform is configured."
    )

    assert (
        OrderedSet([AbbreviatedPlatform.create(current_platform)])
        == Targets(platforms=(current_platform,)).unique_targets()
    )

    assert (
        OrderedSet(LocalInterpreter.create(i) for i in (py27, py38, py310))
        == Targets(interpreters=(py27, py38, py310)).unique_targets()
    )

    complete_platform_current = CompletePlatform.from_interpreter(current_interpreter)
    complete_platform_py27 = CompletePlatform.from_interpreter(py27)
    assert (
        OrderedSet([complete_platform_current, complete_platform_py27])
        == Targets(
            complete_platforms=(complete_platform_current, complete_platform_py27)
        ).unique_targets()
    )


def assert_python_requirement_applies(
    expected_result,  # type: bool
    target,  # type: Target
    python_requirement,  # type: str
    source=None,  # type: Optional[Requirement]
):
    # type: (...) -> None
    assert expected_result == target.requires_python_applies(
        SpecifierSet(python_requirement), source=source or Requirement.parse("foo")
    )


def test_requires_python_current():
    # type: () -> None

    current_target = targets.current()
    major, minor, patch = current_target.interpreter.version
    patch_is_zero = 0 == patch

    def requires_python(template):
        # type: (str) -> str
        return template.format(major=major, minor=minor, patch=patch)

    def assert_requires_python(
        expected_result,  # type: bool
        template,  # type: str
    ):
        # type: (...) -> None
        assert_python_requirement_applies(
            expected_result=expected_result,
            target=current_target,
            python_requirement=requires_python(template),
        )

    assert_requires_python(True, "~={major}.{minor}")

    assert_requires_python(True, "=={major}.{minor}.*")
    assert_requires_python(patch_is_zero, "=={major}.{minor}")
    assert_requires_python(True, "=={major}.{minor}.{patch}")

    assert_requires_python(True, "!={major}")
    assert_requires_python(False, "!={major}.*")
    assert_requires_python(not patch_is_zero, "!={major}.{minor}")
    assert_requires_python(False, "!={major}.{minor}.*")
    assert_requires_python(False, "!={major}.{minor}.{patch}")

    assert_requires_python(False, "<{major}")
    assert_requires_python(False, "<={major}")
    assert_requires_python(False, "<{major}.{minor}")
    assert_requires_python(patch_is_zero, "<={major}.{minor}")
    assert_requires_python(False, "<{major}.{minor}.{patch}")
    assert_requires_python(True, "<={major}.{minor}.{patch}")

    assert_requires_python(True, ">{major}")
    assert_requires_python(True, ">={major}")
    assert_requires_python(not patch_is_zero, ">{major}.{minor}")
    assert_requires_python(True, ">={major}.{minor}")
    assert_requires_python(False, ">{major}.{minor}.{patch}")
    assert_requires_python(True, ">={major}.{minor}.{patch}")

    assert_requires_python(False, "==={major}.{minor}")
    assert_requires_python(True, "==={major}.{minor}.{patch}")


def test_requires_python_abbreviated_platform():
    abbreviated_platform = AbbreviatedPlatform.create(Platform.create("linux-x86_64-cp-37-m"))

    def assert_requires_python(
        expected_result,  # type: bool
        requires_python,  # type: str
    ):
        # type: (...) -> None
        assert_python_requirement_applies(
            expected_result=expected_result,
            target=abbreviated_platform,
            python_requirement=requires_python,
        )

    assert_requires_python(True, "~=3.7")

    assert_requires_python(True, "==3.7.*")
    assert_requires_python(True, "==3.7")
    assert_requires_python(True, "==3.7.0")
    assert_requires_python(False, "==3.7.1")

    assert_requires_python(True, "!=3")
    assert_requires_python(False, "!=3.*")
    assert_requires_python(False, "!=3.7")
    assert_requires_python(False, "!=3.7.*")
    assert_requires_python(False, "!=3.7.0")
    assert_requires_python(True, "!=3.7.1")

    assert_requires_python(False, "<3")
    assert_requires_python(False, "<=3")
    assert_requires_python(False, "<3.7")
    assert_requires_python(True, "<=3.7")
    assert_requires_python(False, "<3.7.0")
    assert_requires_python(True, "<3.7.1")
    assert_requires_python(True, "<=3.7.0")

    assert_requires_python(True, ">3")
    assert_requires_python(True, ">=3")
    assert_requires_python(False, ">3.7")
    assert_requires_python(True, ">=3.7")
    assert_requires_python(False, ">3.7.0")
    assert_requires_python(True, ">=3.7.0")

    assert_requires_python(True, "===3.7")

    # There is no zeo-padding for the `===` operator:
    # https://www.python.org/dev/peps/pep-0440/#arbitrary-equality
    assert_requires_python(False, "===3.7.0")

    assert_requires_python(False, "===3.7.1")


def test_requires_python_invalid_target():
    # type: () -> None

    # This target has an empty marker environment; so no `python_version` and no
    # `python_full_version`.
    invalid_target = CompletePlatform.create(
        marker_environment=MarkerEnvironment(),
        supported_tags=CompatibilityTags.from_strings(["cp37-cp37m-linux_x86_64"]),
    )

    requires_python = SpecifierSet(">=2.7")
    source = Requirement.parse("foo==1.2.3")
    with pytest.raises(
        RequiresPythonError,
        match=r".*{}.*".format(
            re.escape(
                "Encountered `Requires-Python: >=2.7` when evaluating foo==1.2.3 for applicability "
                "but the Python version information needed to evaluate this requirement is not "
                "contained in the target being evaluated for: cp37-cp37m-linux_x86_64"
            )
        ),
    ):
        invalid_target.requires_python_applies(requires_python=requires_python, source=source)


def test_from_target_local_interpreter():
    # type: () -> None

    local_interpreter = LocalInterpreter.create()
    tgts = Targets.from_target(local_interpreter)
    assert local_interpreter.interpreter == tgts.interpreter
    assert OrderedSet([local_interpreter]) == tgts.unique_targets(only_explicit=False)
    assert OrderedSet([local_interpreter]) == tgts.unique_targets(only_explicit=True)


def test_from_target_abbreviated_platform(current_platform):
    # type: (Platform) -> None

    abbreviated_platform = AbbreviatedPlatform.create(
        platform=current_platform, manylinux="manylinux1"
    )
    tgts = Targets.from_target(abbreviated_platform)
    assert tgts.interpreter is None
    assert OrderedSet([abbreviated_platform]) == tgts.unique_targets(only_explicit=False)
    assert OrderedSet([abbreviated_platform]) == tgts.unique_targets(only_explicit=True)


def test_from_target_complete_platform(current_interpreter):
    # type: (PythonInterpreter) -> None

    complete_platform = CompletePlatform.from_interpreter(current_interpreter)
    tgts = Targets.from_target(complete_platform)
    assert tgts.interpreter is None
    assert OrderedSet([complete_platform]) == tgts.unique_targets(only_explicit=False)
    assert OrderedSet([complete_platform]) == tgts.unique_targets(only_explicit=True)



# Copyright 2015 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os
from contextlib import contextmanager
from io import open

from pex.common import temporary_dir
from pex.inherit_path import InheritPath
from pex.pex_builder import PEXBuilder
from pex.testing import run_simple_pex
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterator, Text


@contextmanager
def write_and_run_simple_pex(inheriting):
    # type: (InheritPath.Value) -> Iterator[Text]
    """Write a pex file that contains an executable entry point.

    :param inheriting: whether this pex should inherit site-packages paths.
    """
    with temporary_dir() as td:
        pex_path = os.path.join(td, "show_path.pex")
        with open(os.path.join(td, "exe.py"), "w") as fp:
            fp.write(u"")  # No contents, we just want the startup messages

        pb = PEXBuilder(path=td, preamble=None)
        pb.info.inherit_path = inheriting
        pb.set_executable(os.path.join(td, "exe.py"))
        pb.freeze()
        pb.build(pex_path)
        stdout, _ = run_simple_pex(pex_path, env={"PEX_VERBOSE": "1"})
        yield stdout.decode("utf-8")


def test_inherits_path_fallback_option():
    # type: () -> None
    with write_and_run_simple_pex(inheriting=InheritPath.FALLBACK) as so:
        assert "Scrubbing from user site" not in so, "User packages should not be scrubbed."
        assert "Scrubbing from site-packages" not in so, "Site packages should not be scrubbed."


def test_inherits_path_prefer_option():
    # type: () -> None
    with write_and_run_simple_pex(inheriting=InheritPath.PREFER) as so:
        assert "Scrubbing from user site" not in so, "User packages should not be scrubbed."
        assert "Scrubbing from site-packages" not in so, "Site packages should not be scrubbed."


def test_does_not_inherit_path_option():
    # type: () -> None
    with write_and_run_simple_pex(inheriting=InheritPath.FALSE) as so:
        assert "Scrubbing from user site" in so, "User packages should be scrubbed."
        assert "Scrubbing from site-packages" in so, "Site packages should be scrubbed."



#!/usr/bin/env python3
# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

# N.B.: This script must work with all Pythons we test against including Python 2.7.

from __future__ import print_function

import os
import re


def main():
    # type: () -> None
    print("Test Control Environment Variables:")
    for var, value in sorted(os.environ.items()):
        if re.search(r"(PEX|PYTHON)", var) and var != "PYTHONHASHSEED":
            print("{var}={value}".format(var=var, value=value))


if __name__ == "__main__":
    main()



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

"""The PEX packaging toolchain."""  # N.B.: Flit uses this as our distribution description.

from __future__ import absolute_import

from .version import __version__ as __pex_version__

__version__ = __pex_version__  # N.B.: Flit uses this as out distribution version.



from collections import OrderedDict
if "__PEX_UNVENDORED__" in __import__("os").environ:
  from toml import TomlEncoder  # vendor:skip
else:
  from pex.third_party.toml import TomlEncoder

if "__PEX_UNVENDORED__" in __import__("os").environ:
  from toml import TomlDecoder  # vendor:skip
else:
  from pex.third_party.toml import TomlDecoder



class TomlOrderedDecoder(TomlDecoder):

    def __init__(self):
        super(self.__class__, self).__init__(_dict=OrderedDict)


class TomlOrderedEncoder(TomlEncoder):

    def __init__(self):
        super(self.__class__, self).__init__(_dict=OrderedDict)



import os.path
import subprocess
import sys

from pex.testing import make_env, run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_compile_error_as_warning(
    tmpdir,  # type: Any
):
    # type: (...) -> None
    pex = os.path.join(str(tmpdir), "pex.pex")
    run_pex_command(
        args=[
            "--include-tools",
            "aenum==3.1.11",
            "-o",
            pex,
        ],
        python=sys.executable,
    ).assert_success()
    # The packaged Pex PEX should work with all Pythons we support, including the current test
    # interpreter.
    python_version = sys.version_info[0]
    compiled_deps_dir = str(tmpdir)
    process = subprocess.Popen(
        args=[sys.executable, pex, "venv", "--scope", "deps", "--compile", compiled_deps_dir],
        env=make_env(PEX_PYTHON=sys.executable, PEX_TOOLS=1),
        stderr=subprocess.PIPE,
    )
    _, stderr_bytes = process.communicate()
    assert 0 == process.returncode

    stderr_str = stderr_bytes.decode("utf-8")
    if python_version == 2:
        assert not stderr_str, "No PEXWarning should be generated for python2"
    else:
        assert (
            "PEXWarning: ignoring compile error" in stderr_str
        ), "PEXWarning should be generated for python3."



# coding=utf-8
# Copyright 2018 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import os


class Bootstrap(object):
    """Supports introspection of the PEX bootstrap code."""

    _INSTANCE = None

    @classmethod
    def locate(cls):
        # type: () -> Bootstrap
        """Locates the active PEX bootstrap.

        :rtype: :class:`Bootstrap`
        """
        if cls._INSTANCE is None:
            bootstrap_path = __file__
            module_import_path = __name__.split(".")

            # For example, our __file__ might be requests.pex/.bootstrap/pex/bootstrap.pyc and our import
            # path pex.bootstrap; so we walk back through all the module components of our import path to
            # find the base sys.path entry where we were found (requests.pex/.bootstrap in this example).
            for _ in module_import_path:
                bootstrap_path = os.path.dirname(bootstrap_path)

            cls._INSTANCE = cls(sys_path_entry=bootstrap_path)
        return cls._INSTANCE

    def __init__(self, sys_path_entry):
        # type: (str) -> None
        self._sys_path_entry = sys_path_entry
        self._realpath = os.path.realpath(self._sys_path_entry)

    @property
    def path(self):
        # type: () -> str
        return self._sys_path_entry

    def demote(self):
        """Demote the bootstrap code to the end of the `sys.path` so it is found last.

        :return: The list of un-imported bootstrap modules.
        :rtype: list of :class:`types.ModuleType`
        """
        import sys  # Grab a hold of `sys` early since we'll be un-importing our module in this process.

        # N.B.: We mutate the sys.path before un-importing modules so that any re-imports triggered
        # by concurrent code will pull from the desired sys.path ordering.
        # See here for how this situation might arise: https://github.com/pantsbuild/pex/issues/1272

        sys.path[:] = [path for path in sys.path if os.path.realpath(path) != self._realpath]
        sys.path.append(self._sys_path_entry)

        unimported_modules = []
        for name, module in reversed(sorted(sys.modules.items())):
            if self.imported_from_bootstrap(module):
                unimported_modules.append(sys.modules.pop(name))

        return unimported_modules

    def imported_from_bootstrap(self, module):
        """Return ``True`` if the given ``module`` object was imported from bootstrap code.

        :param module: The module to check the provenance of.
        :type module: :class:`types.ModuleType`
        :rtype: bool
        """

        # A vendored module.
        path = getattr(module, "__file__", None)
        if path and os.path.realpath(path).startswith(self._realpath):
            return True

        # A vendored package.
        path = getattr(module, "__path__", None)
        if path and any(
            os.path.realpath(path_item).startswith(self._realpath) for path_item in path
        ):
            return True

        return False

    def __repr__(self):
        return "{cls}(sys_path_entry={sys_path_entry!r})".format(
            cls=type(self).__name__, sys_path_entry=self._sys_path_entry
        )



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import ast
import os

from pex.common import is_python_script
from pex.dist_metadata import Distribution, EntryPoint
from pex.pep_376 import InstalledWheel
from pex.pep_503 import ProjectName
from pex.typing import TYPE_CHECKING, cast

if TYPE_CHECKING:
    from typing import Dict, Iterable, Optional, Tuple

    import attr  # vendor:skip
else:
    from pex.third_party import attr


@attr.s(frozen=True)
class DistributionScript(object):
    @classmethod
    def find(
        cls,
        dist,  # type: Distribution
        name,  # type: str
    ):
        # type: (...) -> Optional[DistributionScript]
        script_path = InstalledWheel.load(dist.location).stashed_path("bin", name)
        return cls(dist=dist, path=script_path) if os.path.isfile(script_path) else None

    dist = attr.ib()  # type: Distribution
    path = attr.ib()  # type: str

    def read_contents(self):
        # type: () -> bytes
        with open(self.path, "rb") as fp:
            return fp.read()

    def python_script(self):
        # type: () -> Optional[ast.AST]
        if not is_python_script(self.path):
            return None

        try:
            return cast(
                ast.AST, compile(self.read_contents(), self.path, "exec", flags=0, dont_inherit=1)
            )
        except (SyntaxError, TypeError):
            return None


def get_script_from_distributions(
    name,  # type: str
    dists,  # type: Iterable[Distribution]
):
    # type: (...) -> Optional[DistributionScript]
    for dist in dists:
        distribution_script = DistributionScript.find(dist, name)
        if distribution_script:
            return distribution_script
    return None


@attr.s(frozen=True)
class DistributionEntryPoint(object):
    dist = attr.ib()  # type: Distribution
    entry_point = attr.ib()  # type: EntryPoint


def get_entry_point_from_console_script(
    script,  # type: str
    dists,  # type: Iterable[Distribution]
):
    # type: (...) -> Optional[DistributionEntryPoint]
    # Check all distributions for the console_script "script". De-dup by dist key to allow for a
    # duplicate console script IFF the distribution is platform-specific and this is a
    # multi-platform pex.
    def get_entrypoint(dist):
        # type: (Distribution) -> Optional[EntryPoint]
        return dist.get_entry_map().get("console_scripts", {}).get(script)

    entries = {}  # type: Dict[ProjectName, DistributionEntryPoint]
    for dist in dists:
        entry_point = get_entrypoint(dist)
        if entry_point is not None:
            entries[dist.metadata.project_name] = DistributionEntryPoint(dist, entry_point)

    if len(entries) > 1:
        raise RuntimeError(
            "Ambiguous script specification {script} matches multiple entry points:\n\t"
            "{entry_points}".format(
                script=script,
                entry_points="\n\t".join(
                    "{entry_point} from {dist}".format(
                        entry_point=dist_entry_point.entry_point, dist=dist_entry_point.dist
                    )
                    for dist_entry_point in entries.values()
                ),
            )
        )

    dist_entry_point = None
    if entries:
        dist_entry_point = next(iter(entries.values()))
    return dist_entry_point



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import json
import os.path
import sys
from argparse import ArgumentTypeError, Namespace, _ActionsContainer

from pex.argparse import HandleBoolAction
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import InterpreterConstraints
from pex.orderedset import OrderedSet
from pex.pep_425 import CompatibilityTags
from pex.pep_508 import MarkerEnvironment
from pex.platforms import Platform
from pex.resolve.resolver_options import _ManylinuxAction
from pex.resolve.target_configuration import InterpreterConfiguration, TargetConfiguration
from pex.targets import CompletePlatform


def register(
    parser,  # type: _ActionsContainer
    include_platforms=True,  # type: bool
):
    # type: (...) -> None
    """Register resolve target selection options with the given parser.

    :param parser: The parser to register target selection options with.
    :param include_platforms: Whether to include options to select targets by platform.
    """

    parser.add_argument(
        "--python",
        dest="python",
        default=[],
        type=str,
        action="append",
        help=(
            "The Python interpreter to use (default: current interpreter). Either specify an "
            "absolute path to an interpreter, or specify a binary accessible on $PATH like "
            "`python3.7`. This option can be passed multiple times."
        ),
    )
    parser.add_argument(
        "--python-path",
        dest="python_path",
        default=None,
        type=str,
        help=(
            "A {pathsep!r} separated list of paths to search for interpreters in when "
            "`--interpreter-constraint` {and_maybe_platforms} specified (default: $PATH). Each "
            "element can be the absolute path of an interpreter binary or a directory containing "
            "interpreter binaries.".format(
                pathsep=os.pathsep,
                and_maybe_platforms=(
                    "and/or `--resolve-local-platforms` are" if include_platforms else "is"
                ),
            )
        ),
    )

    current_interpreter = PythonInterpreter.get()
    program = sys.argv[0]
    singe_interpreter_info_cmd = (
        "PEX_TOOLS=1 {current_interpreter} {program} interpreter --verbose --indent 4".format(
            current_interpreter=current_interpreter.binary, program=program
        )
    )
    all_interpreters_info_cmd = (
        "PEX_TOOLS=1 {program} interpreter --all --verbose --indent 4".format(program=program)
    )

    parser.add_argument(
        "--interpreter-constraint",
        dest="interpreter_constraint",
        default=[],
        type=str,
        action="append",
        help=(
            "Constrain the selected Python interpreter. Specify with Requirement-style syntax, "
            'e.g. "CPython>=2.7,<3" (A CPython interpreter with version >=2.7 AND version <3), '
            '">=2.7,<3" (Any Python interpreter with version >=2.7 AND version <3) or "PyPy" (A '
            "PyPy interpreter of any version). This argument may be repeated multiple times to OR "
            "the constraints. Try `{singe_interpreter_info_cmd}` to find the exact interpreter "
            "constraints of {current_interpreter} and `{all_interpreters_info_cmd}` to find out "
            "the interpreter constraints of all Python interpreters on the $PATH.".format(
                current_interpreter=current_interpreter.binary,
                singe_interpreter_info_cmd=singe_interpreter_info_cmd,
                all_interpreters_info_cmd=all_interpreters_info_cmd,
            )
        ),
    )

    if include_platforms:
        _register_platform_options(
            parser, current_interpreter, singe_interpreter_info_cmd, all_interpreters_info_cmd
        )


def _register_platform_options(
    parser,  # type: _ActionsContainer
    current_interpreter,  # type: PythonInterpreter
    singe_interpreter_info_cmd,  # type: str
    all_interpreters_info_cmd,  # type: str
):
    # type: (...) -> None
    parser.add_argument(
        "--platform",
        "--abbreviated-platform",
        dest="platforms",
        default=[],
        type=str,
        action="append",
        help=(
            "The (abbreviated) platform to build the PEX for. This option can be passed multiple "
            "times to create a multi-platform pex. To use the platform corresponding to the "
            "current interpreter you can pass `current`. To target any other platform you pass a "
            "string composed of fields: <platform>-<python impl abbr>-<python version>-<abi>. "
            "These fields stem from wheel name conventions as outlined in "
            "https://www.python.org/dev/peps/pep-0427#file-name-convention and influenced by "
            "https://www.python.org/dev/peps/pep-0425. For the current interpreter at "
            "{current_interpreter} the full platform string is {current_platform}. To find out "
            "more, try `{all_interpreters_info_cmd}` to print out the platform for all "
            "interpreters on the $PATH or `{singe_interpreter_info_cmd}` to inspect the single "
            "interpreter {current_interpreter}.".format(
                current_interpreter=current_interpreter.binary,
                current_platform=current_interpreter.platform,
                singe_interpreter_info_cmd=singe_interpreter_info_cmd,
                all_interpreters_info_cmd=all_interpreters_info_cmd,
            )
        ),
    )

    parser.add_argument(
        "--complete-platform",
        dest="complete_platforms",
        default=[],
        type=str,
        action="append",
        help=(
            "The complete platform information describing the platform for which to build the PEX. "
            "This option can be passed multiple times to create a multi-platform pex. Values "
            "should be either JSON object literal strings or paths to files containing them. The "
            "JSON object is expected to have two fields with any other fields ignored. The "
            "'marker_environment' field should have an object value with string field values "
            "corresponding to PEP-508 marker environment entries (See: "
            "https://www.python.org/dev/peps/pep-0508/#environment-markers). It is OK to only have "
            "a subset of valid marker environment fields but it is not valid to present entries "
            "not defined in PEP-508. The 'compatible_tags' field should have an array of strings "
            "value containing the compatible tags in order from most specific first to least "
            "specific last as defined in PEP-425 (See: https://www.python.org/dev/peps/pep-0425). "
            "Pex can create complete platform JSON for you by running it on the target platform "
            "like so: `pex3 interpreter inspect --markers --tags`. For more options, particularly "
            "to select the desired target interpreter see: `pex3 interpreter inspect --help`."
        ),
    )

    default_target_configuration = TargetConfiguration()
    parser.add_argument(
        "--manylinux",
        "--no-manylinux",
        "--no-use-manylinux",
        dest="assume_manylinux",
        type=str,
        default=default_target_configuration.assume_manylinux,
        action=_ManylinuxAction,
        help="Whether to allow resolution of manylinux wheels for linux target platforms.",
    )

    parser.add_argument(
        "--resolve-local-platforms",
        dest="resolve_local_platforms",
        default=False,
        action=HandleBoolAction,
        help=(
            "When --platforms are specified, attempt to resolve a local interpreter that matches "
            "each platform specified. If found, use the interpreter to resolve distributions; if "
            "not (or if this option is not specified), resolve for each platform only allowing "
            "matching binary distributions and failing if only sdists or non-matching binary "
            "distributions can be found."
        ),
    )


def configure_interpreters(options):
    # type: (Namespace) -> InterpreterConfiguration
    """Creates an interpreter configuration from options registered by `register`.

    :param options: The interpreter configuration options.
    """
    try:
        interpreter_constraints = InterpreterConstraints.parse(*options.interpreter_constraint)
    except ValueError as e:
        raise ArgumentTypeError(str(e))

    return InterpreterConfiguration(
        interpreter_constraints=interpreter_constraints,
        python_path=(
            tuple(OrderedSet(options.python_path.split(os.pathsep)))
            if options.python_path
            else None
        ),
        pythons=tuple(OrderedSet(options.python)),
    )


def _create_complete_platform(value):
    # type: (str) -> CompletePlatform
    if os.path.isfile(value):
        try:
            with open(value) as fp:
                data = json.load(fp)
        except (OSError, ValueError) as e:
            raise ArgumentTypeError(
                "Failed to load complete platform data from {path}: {err}".format(path=value, err=e)
            )
    else:
        try:
            data = json.loads(value)
        except ValueError as e:
            raise ArgumentTypeError(
                "Failed to load complete platform data from json string: {err}".format(err=e)
            )

    try:
        marker_environment = MarkerEnvironment(**data["marker_environment"])
    except KeyError:
        raise ArgumentTypeError(
            "The complete platform JSON object did not have the required 'marker_environment' "
            "key:\n{json_object}".format(json_object=json.dumps(data, indent=4))
        )
    except TypeError as e:
        raise ArgumentTypeError(
            "Invalid environment entry provided: {err}\n"
            "See https://www.python.org/dev/peps/pep-0508/#environment-markers for valid "
            "entries.".format(err=e)
        )

    try:
        supported_tags = CompatibilityTags.from_strings(data["compatible_tags"])
    except KeyError:
        raise ArgumentTypeError(
            "The complete platform JSON object did not have the required 'compatible_tags' "
            "key:\n{json_object}".format(json_object=json.dumps(data, indent=4))
        )

    return CompletePlatform.create(marker_environment, supported_tags)


def configure(options):
    # type: (Namespace) -> TargetConfiguration
    """Creates a target configuration from options via `register(..., include_platforms=True)`.

    :param options: The target configuration options.
    """
    interpreter_configuration = configure_interpreters(options)

    try:
        platforms = tuple(
            OrderedSet(
                Platform.create(platform) if platform and platform != "current" else None
                for platform in options.platforms
            )
        )
    except Platform.InvalidPlatformError as e:
        raise ArgumentTypeError(str(e))

    complete_platforms = tuple(
        OrderedSet(_create_complete_platform(value) for value in options.complete_platforms)
    )

    return TargetConfiguration(
        interpreter_configuration=interpreter_configuration,
        platforms=platforms,
        complete_platforms=complete_platforms,
        resolve_local_platforms=options.resolve_local_platforms,
        assume_manylinux=options.assume_manylinux,
    )



# Copyright 2014 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

import hashlib
import os
import sys

from pex import pex_warnings
from pex.atomic_directory import atomic_directory
from pex.common import die, pluralize
from pex.environment import ResolveError
from pex.inherit_path import InheritPath
from pex.interpreter import PythonInterpreter
from pex.interpreter_constraints import (
    InterpreterConstraints,
    UnsatisfiableInterpreterConstraintsError,
)
from pex.layout import Layout
from pex.orderedset import OrderedSet
from pex.pex_info import PexInfo
from pex.targets import LocalInterpreter
from pex.tracer import TRACER
from pex.typing import TYPE_CHECKING, cast
from pex.variables import ENV
from pex.venv import installer

if TYPE_CHECKING:
    from typing import Iterable, Iterator, List, NoReturn, Optional, Set, Tuple, Union

    import attr  # vendor:skip

    from pex.interpreter import InterpreterIdentificationError, InterpreterOrError, PathFilter
    from pex.pex import PEX
else:
    from pex.third_party import attr


def normalize_path(path):
    # type: (Optional[Iterable[str]]) -> Optional[OrderedSet[str]]
    """Normalizes a PATH list into a de-duped list of paths."""
    return OrderedSet(PythonInterpreter.canonicalize_path(p) for p in path) if path else None


@attr.s(frozen=True)
class InterpreterTest(object):
    entry_point = attr.ib()  # type: str
    pex_info = attr.ib()  # type: PexInfo

    @property
    def interpreter_constraints(self):
        # type: () -> InterpreterConstraints
        return self.pex_info.interpreter_constraints

    def test_resolve(self, interpreter):
        # type: (PythonInterpreter) -> Union[ResolveError, bool]
        """Checks if `interpreter` can resolve all required distributions for the PEX under test."""
        with TRACER.timed(
            "Testing {python} can resolve PEX at {pex}".format(
                python=interpreter.binary, pex=self.entry_point
            )
        ):
            from pex.environment import PEXEnvironment

            pex_environment = PEXEnvironment.mount(
                self.entry_point,
                pex_info=self.pex_info,
                target=LocalInterpreter.create(interpreter),
            )
            try:
                pex_environment.resolve()
                return True
            except ResolveError as e:
                return e


# TODO(John Sirois): Move this to interpreter_constraints.py. As things stand, both pex/bin/pex.py
#  and this file use this function. The Pex CLI should not depend on this file which hosts code
#  used at PEX runtime.
def iter_compatible_interpreters(
    path=None,  # type: Optional[Iterable[str]]
    valid_basenames=None,  # type: Optional[Iterable[str]]
    interpreter_constraints=None,  # type: Optional[InterpreterConstraints]
    preferred_interpreter=None,  # type: Optional[PythonInterpreter]
    interpreter_test=None,  # type: Optional[InterpreterTest]
):
    # type: (...) -> Iterator[PythonInterpreter]
    """Find all compatible interpreters on the system within the supplied constraints.

    :param path: A search PATH of files or directories.
    :param valid_basenames: Valid basenames for discovered interpreter binaries. If not specified,
                            Then all typical names are accepted (i.e.: python, python3, python2.7,
                            pypy, etc.).
    :param interpreter_constraints: Interpreter type and version constraint strings as described in
                                    `--interpreter-constraint`.
    :param preferred_interpreter: For testing - an interpreter to prefer amongst all others.
                                  Defaults to the current running interpreter.
    :param interpreter_test: Optional test to verify selected interpreters can boot a given PEX.
    Interpreters are searched for in `path` if specified and $PATH if not.

    If no interpreters are found and there are no further constraints (neither `valid_basenames` nor
    `interpreter_constraints` is specified) then the returned iterator will be empty. However, if
    there are constraints specified, the returned iterator, although empty, will raise
    `UnsatisfiableInterpreterConstraintsError` to provide information about any found interpreters
    that did not match all the constraints.
    """

    _valid_path = None  # type: Optional[PathFilter]
    if valid_basenames:
        _valid_basenames = frozenset(cast("Iterable[str]", valid_basenames))
        _valid_path = (
            lambda interpreter_path: os.path.basename(interpreter_path) in _valid_basenames
        )

    def _iter_interpreters():
        # type: () -> Iterator[InterpreterOrError]
        seen = set()  # type: Set[InterpreterOrError]

        normalized_paths = normalize_path(path)

        # Prefer the current interpreter, if valid.
        current_interpreter = preferred_interpreter or PythonInterpreter.get()
        if not _valid_path or _valid_path(current_interpreter.binary):
            if normalized_paths:
                candidate_paths = frozenset(
                    (current_interpreter.binary, os.path.dirname(current_interpreter.binary))
                )
                candidate_paths_in_path = candidate_paths.intersection(normalized_paths)
                if candidate_paths_in_path:
                    # In case the full path of the current interpreter binary was in the
                    # `normalized_paths` we're searching, remove it to prevent identifying it again
                    # just to then skip it as `seen`.
                    normalized_paths.discard(current_interpreter.binary)
                    seen.add(current_interpreter)
                    yield current_interpreter
            else:
                seen.add(current_interpreter)
                yield current_interpreter

        for interp in PythonInterpreter.iter_candidates(
            paths=normalized_paths, path_filter=_valid_path
        ):
            if interp not in seen:
                seen.add(interp)
                yield interp

    def _valid_interpreter(interp):
        # type: (PythonInterpreter) -> Union[ResolveError, bool]
        if not interpreter_constraints:
            return interpreter_test.test_resolve(interp) if interpreter_test else True

        if interp in interpreter_constraints:
            TRACER.log(
                "Constraints on interpreters: {}, Matching Interpreter: {}".format(
                    interpreter_constraints, interp.binary
                ),
                V=3,
            )
            return interpreter_test.test_resolve(interp) if interpreter_test else True

        return False

    candidates = []  # type: List[PythonInterpreter]
    resolve_errors = []  # type: List[ResolveError]
    identification_failures = []  # type: List[InterpreterIdentificationError]
    found = False

    for interpreter_or_error in _iter_interpreters():
        if isinstance(interpreter_or_error, PythonInterpreter):
            interpreter = interpreter_or_error
            candidates.append(interpreter)
            valid_or_error = _valid_interpreter(interpreter)
            if isinstance(valid_or_error, ResolveError):
                resolve_errors.append(valid_or_error)
            elif valid_or_error:
                found = True
                yield interpreter
        else:
            identification_failures.append(interpreter_or_error)

    if not found and (resolve_errors or interpreter_constraints or valid_basenames):
        constraints = []  # type: List[str]
        if resolve_errors:
            constraints.extend(str(resolve_error) for resolve_error in resolve_errors)
        else:
            if interpreter_constraints:
                constraints.append(
                    "Version matches {}".format(" or ".join(map(str, interpreter_constraints)))
                )
            if valid_basenames:
                constraints.append("Basename is {}".format(" or ".join(valid_basenames)))
        raise UnsatisfiableInterpreterConstraintsError(
            constraints, candidates, identification_failures
        )


def _select_path_interpreter(
    path=None,  # type: Optional[Tuple[str, ...]]
    valid_basenames=None,  # type: Optional[Tuple[str, ...]]
    interpreter_constraints=None,  # type: Optional[InterpreterConstraints]
    preferred_interpreter=None,  # type: Optional[PythonInterpreter]
    interpreter_test=None,  # type: Optional[InterpreterTest]
):
    # type: (...) -> Optional[PythonInterpreter]

    candidate_interpreters_iter = iter_compatible_interpreters(
        path=path,
        valid_basenames=valid_basenames,
        interpreter_constraints=interpreter_constraints,
        preferred_interpreter=preferred_interpreter,
        interpreter_test=interpreter_test,
    )
    current_interpreter = PythonInterpreter.get()  # type: PythonInterpreter
    preferred_interpreter = preferred_interpreter or current_interpreter
    candidate_interpreters = OrderedSet()  # type: OrderedSet[PythonInterpreter]
    for interpreter in candidate_interpreters_iter:
        if preferred_interpreter == interpreter:
            # Always respect the preferred interpreter if it doesn't violate other constraints.
            return preferred_interpreter
        else:
            candidate_interpreters.add(interpreter)
    if not candidate_interpreters:
        return None
    if current_interpreter in candidate_interpreters:
        # Always prefer continuing with the current interpreter when possible to avoid re-exec
        # overhead.
        return current_interpreter
    # TODO: Allow the selection strategy to be parameterized:
    #   https://github.com/pantsbuild/pex/issues/430
    return PythonInterpreter.latest_release_of_min_compatible_version(candidate_interpreters)


def find_compatible_interpreter(interpreter_test=None):
    # type: (Optional[InterpreterTest]) -> PythonInterpreter

    interpreter_constraints = interpreter_test.interpreter_constraints if interpreter_test else None

    def gather_constraints():
        # type: () -> Iterable[str]
        constraints = []
        if ENV.PEX_PYTHON:
            constraints.append("PEX_PYTHON={}".format(ENV.PEX_PYTHON))
        if ENV.PEX_PYTHON_PATH:
            constraints.append("PEX_PYTHON_PATH={}".format(ENV.PEX_PYTHON_PATH))
        if interpreter_constraints:
            constraints.append("Version matches {}".format(interpreter_constraints))
        return constraints

    preferred_interpreter = None  # type: Optional[PythonInterpreter]
    if ENV.PEX_PYTHON and os.path.isabs(ENV.PEX_PYTHON):
        try:
            preferred_interpreter = PythonInterpreter.from_binary(ENV.PEX_PYTHON)
        except PythonInterpreter.Error as e:
            raise UnsatisfiableInterpreterConstraintsError(
                constraints=gather_constraints(),
                candidates=[],
                failures=[(ENV.PEX_PYTHON, str(e))],
                preamble=(
                    "The specified PEX_PYTHON={pex_python} could not be identified as a "
                    "valid Python interpreter.".format(pex_python=ENV.PEX_PYTHON)
                ),
            )

    current_interpreter = PythonInterpreter.get()
    with TRACER.timed("Selecting runtime interpreter", V=3):
        if ENV.PEX_PYTHON and not ENV.PEX_PYTHON_PATH:
            TRACER.log(
                "Using PEX_PYTHON={} constrained by {}".format(
                    ENV.PEX_PYTHON, interpreter_constraints
                ),
                V=3,
            )
            try:
                if os.path.isabs(ENV.PEX_PYTHON):
                    target = _select_path_interpreter(
                        path=(ENV.PEX_PYTHON,),
                        interpreter_constraints=interpreter_constraints,
                        preferred_interpreter=preferred_interpreter,
                        interpreter_test=interpreter_test,
                    )
                else:
                    target = _select_path_interpreter(
                        valid_basenames=(os.path.basename(ENV.PEX_PYTHON),),
                        interpreter_constraints=interpreter_constraints,
                        interpreter_test=interpreter_test,
                    )
            except UnsatisfiableInterpreterConstraintsError as e:
                raise e.with_preamble(
                    "Failed to find a compatible PEX_PYTHON={pex_python}.".format(
                        pex_python=ENV.PEX_PYTHON
                    )
                )
        else:
            TRACER.log(
                "Using {path} constrained by {constraints}".format(
                    path="PEX_PYTHON_PATH={}".format(ENV.PEX_PYTHON_PATH)
                    if ENV.PEX_PYTHON_PATH
                    else "$PATH",
                    constraints=interpreter_constraints,
                ),
                V=3,
            )
            try:
                target = _select_path_interpreter(
                    path=ENV.PEX_PYTHON_PATH,
                    interpreter_constraints=interpreter_constraints,
                    preferred_interpreter=preferred_interpreter,
                    interpreter_test=interpreter_test,
                )
            except UnsatisfiableInterpreterConstraintsError as e:
                raise e.with_preamble(
                    "Failed to find compatible interpreter on path {path}.".format(
                        path=(
                            os.pathsep.join(ENV.PEX_PYTHON_PATH)
                            if ENV.PEX_PYTHON_PATH
                            else os.getenv("PATH")
                        )
                    )
                )

        if preferred_interpreter and target != preferred_interpreter:
            candidates = [preferred_interpreter, target] if target else [preferred_interpreter]
            raise UnsatisfiableInterpreterConstraintsError(
                constraints=gather_constraints(),
                candidates=candidates,
                failures=[],
                preamble=(
                    "The specified PEX_PYTHON={pex_python} did not meet other "
                    "constraints.".format(pex_python=ENV.PEX_PYTHON)
                ),
            )

        if target is None:
            # N.B.: This can only happen when PEX_PYTHON_PATH is set and interpreter_constraints
            # is empty, but we handle all constraints generally for sanity sake.
            raise UnsatisfiableInterpreterConstraintsError(
                constraints=gather_constraints(),
                candidates=[current_interpreter],
                failures=[],
                preamble="Could not find a compatible interpreter.",
            )

        return target


def maybe_reexec_pex(interpreter_test):
    # type: (InterpreterTest) -> Union[None, NoReturn]
    """Handle environment overrides for the Python interpreter to use when executing this pex.

    This function supports interpreter filtering based on interpreter constraints stored in PEX-INFO
    metadata. If PEX_PYTHON is set it attempts to obtain the binary location of the interpreter
    specified by PEX_PYTHON. If PEX_PYTHON_PATH is set, it attempts to search the path for a
    matching interpreter in accordance with the interpreter constraints. If both variables are
    present, this function gives precedence to PEX_PYTHON_PATH and errors out if no compatible
    interpreters can be found on said path.

    If neither variable is set, we fall back to plain PEX execution using PATH searching or the
    currently executing interpreter. If compatibility constraints are used, we match those
    constraints against these interpreters.

    :param interpreter_test: Optional test to verify selected interpreters can boot a given PEX.
    """

    current_interpreter = PythonInterpreter.get()

    # NB: Used only for tests.
    if "_PEX_EXEC_CHAIN" in os.environ:
        flag_or_chain = os.environ.pop("_PEX_EXEC_CHAIN")
        pex_exec_chain = [] if flag_or_chain == "1" else flag_or_chain.split(os.pathsep)
        pex_exec_chain.append(current_interpreter.binary)
        os.environ["_PEX_EXEC_CHAIN"] = os.pathsep.join(pex_exec_chain)

    current_interpreter_blessed_env_var = "_PEX_SHOULD_EXIT_BOOTSTRAP_REEXEC"
    if os.environ.pop(current_interpreter_blessed_env_var, None):
        # We've already been here and selected an interpreter. Continue to execution.
        return None

    try:
        target = find_compatible_interpreter(interpreter_test=interpreter_test)
    except UnsatisfiableInterpreterConstraintsError as e:
        die(str(e))

    os.environ.pop("PEX_PYTHON", None)
    os.environ.pop("PEX_PYTHON_PATH", None)

    if interpreter_test.pex_info.inherit_path == InheritPath.FALSE:
        # Now that we've found a compatible Python interpreter, make sure we resolve out of any
        # virtual environments it may be contained in since virtual environments created with
        # `--system-site-packages` foil PEX attempts to scrub the sys.path.
        resolved = target.resolve_base_interpreter()
        if resolved != target:
            TRACER.log(
                "Resolved base interpreter of {} from virtual environment at {}".format(
                    resolved, target.prefix
                ),
                V=3,
            )
        target = resolved

    from . import pex

    pythonpath = pex.PEX.stash_pythonpath()
    if pythonpath is not None:
        TRACER.log("Stashed PYTHONPATH of {}".format(pythonpath), V=2)
    elif target == current_interpreter:
        TRACER.log(
            "Using the current interpreter {} since it matches constraints and "
            "PYTHONPATH is not set.".format(sys.executable)
        )
        return None

    target_binary = target.binary
    cmdline = [target_binary] + sys.argv
    TRACER.log(
        "Re-executing: "
        "cmdline={cmdline!r}, "
        "sys.executable={python!r}, "
        "PEX_PYTHON={pex_python!r}, "
        "PEX_PYTHON_PATH={pex_python_path!r}, "
        "interpreter_constraints={interpreter_constraints!r}"
        "{pythonpath}".format(
            cmdline=" ".join(cmdline),
            python=sys.executable,
            pex_python=ENV.PEX_PYTHON,
            pex_python_path=ENV.PEX_PYTHON_PATH,
            interpreter_constraints=interpreter_test.interpreter_constraints,
            pythonpath=', (stashed) PYTHONPATH="{}"'.format(pythonpath)
            if pythonpath is not None
            else "",
        )
    )

    # Avoid a re-run through compatibility_constraint checking.
    os.environ[current_interpreter_blessed_env_var] = "1"

    os.execv(target_binary, cmdline)


def _bootstrap(entry_point):
    # type: (str) -> PexInfo
    pex_info = PexInfo.from_pex(entry_point)  # type: PexInfo
    pex_info.update(PexInfo.from_env())
    pex_warnings.configure_warnings(ENV, pex_info=pex_info)
    return pex_info


@attr.s(frozen=True)
class VenvPex(object):
    venv_dir = attr.ib()  # type: str
    hermetic_scripts = attr.ib(default=True)  # type: bool
    pex = attr.ib(init=False)  # type: str
    python = attr.ib(init=False)  # type: str

    def bin_file(self, name):
        # type: (str) -> str
        return os.path.join(self.venv_dir, "bin", name)

    def __attrs_post_init__(self):
        # type: () -> None
        object.__setattr__(self, "pex", os.path.join(self.venv_dir, "pex"))
        object.__setattr__(self, "python", self.bin_file("python"))

    def execute_args(self, *additional_args):
        # type: (*str) -> List[str]
        argv = [self.python]
        if self.hermetic_scripts:
            argv.append("-sE")
        argv.append(self.pex)
        argv.extend(additional_args)
        return argv

    def execv(self, *additional_args):
        # type: (*str) -> NoReturn
        os.execv(self.python, self.execute_args(*additional_args))


def ensure_venv(
    pex,  # type: PEX
    collisions_ok=True,  # type: bool
):
    # type: (...) -> VenvPex
    pex_info = pex.pex_info()
    venv_dir = pex_info.runtime_venv_dir(pex_file=pex.path(), interpreter=pex.interpreter)
    if venv_dir is None:
        raise AssertionError(
            "Expected PEX-INFO for {} to have the components of a venv directory".format(pex.path())
        )
    if not pex_info.includes_tools:
        raise ValueError(
            "The PEX_VENV environment variable was set, but this PEX was not built with venv "
            "support (Re-build the PEX file with `pex --venv ...`)"
        )
    with atomic_directory(venv_dir) as venv:
        if not venv.is_finalized():
            from pex.venv.virtualenv import Virtualenv

            virtualenv = Virtualenv.create_atomic(
                venv_dir=venv,
                interpreter=pex.interpreter,
                copies=pex_info.venv_copies,
                prompt=os.path.basename(ENV.PEX) if ENV.PEX else None,
            )

            pex_path = os.path.abspath(pex.path())

            # A sha1 hash is 160 bits -> 20 bytes -> 40 hex characters. We start with 8 characters
            # (32 bits) of entropy since that is short and _very_ unlikely to collide with another
            # PEX venv on this machine. If we still collide after using the whole sha1 (for a total
            # of 33 collisions), then the universe is broken and we raise. It's the least we can do.
            venv_hash = hashlib.sha1(venv_dir.encode("utf-8")).hexdigest()
            collisions = []
            for chars in range(8, len(venv_hash) + 1):
                entropy = venv_hash[:chars]
                short_venv_dir = os.path.join(pex_info.pex_root, "venvs", "s", entropy)
                with atomic_directory(short_venv_dir) as short_venv:
                    if short_venv.is_finalized():
                        collisions.append(short_venv_dir)
                        if entropy == venv_hash:
                            raise RuntimeError(
                                "The venv for {pex} at {venv} has hash collisions with {count} "
                                "other {venvs}!\n{collisions}".format(
                                    pex=pex_path,
                                    venv=venv_dir,
                                    count=len(collisions),
                                    venvs=pluralize(collisions, "venv"),
                                    collisions="\n".join(
                                        "{index}.) {venv_path}".format(
                                            index=index, venv_path=os.path.realpath(path)
                                        )
                                        for index, path in enumerate(collisions, start=1)
                                    ),
                                )
                            )
                        continue

                    os.symlink(venv_dir, os.path.join(short_venv.work_dir, "venv"))

                    # Loose PEXes don't need to unpack themselves to the PEX_ROOT before running;
                    # so we'll not have a stable base there to symlink from. As such, always copy
                    # for loose PEXes to ensure the PEX_ROOT venv is stable in the face of
                    # modification of the source loose PEX.
                    symlink = pex.layout != Layout.LOOSE and not pex_info.venv_site_packages_copies

                    shebang = installer.populate_venv_from_pex(
                        virtualenv,
                        pex,
                        bin_path=pex_info.venv_bin_path,
                        python=os.path.join(
                            short_venv_dir,
                            "venv",
                            "bin",
                            os.path.basename(virtualenv.interpreter.binary),
                        ),
                        collisions_ok=collisions_ok,
                        symlink=symlink,
                        hermetic_scripts=pex_info.venv_hermetic_scripts,
                    )

                    # There are popular Linux distributions with shebang length limits
                    # (BINPRM_BUF_SIZE in /usr/include/linux/binfmts.h) set at 128 characters, so
                    # we warn in the _very_ unlikely case that our shortened shebang is longer than
                    # this.
                    if len(shebang) > 128:
                        pex_warnings.warn(
                            "The venv for {pex} at {venv} has script shebangs of {shebang!r} with "
                            "{count} characters. On some systems this may be too long and cause "
                            "problems running the venv scripts. You may be able adjust PEX_ROOT "
                            "from {pex_root} to a shorter path as a work-around.".format(
                                pex=pex_path,
                                venv=venv_dir,
                                shebang=shebang,
                                count=len(shebang),
                                pex_root=pex_info.pex_root,
                            )
                        )

                    break

    return VenvPex(venv_dir, hermetic_scripts=pex_info.venv_hermetic_scripts)


# NB: This helper is used by the PEX bootstrap __main__.py as well as the __pex__/__init__.py
# import hook code.
def bootstrap_pex(
    entry_point,  # type: str
    execute=True,  # type: bool
    venv_dir=None,  # type: Optional[str]
):
    # type: (...) -> None

    pex_info = _bootstrap(entry_point)

    # ENV.PEX_ROOT is consulted by PythonInterpreter and Platform so set that up as early as
    # possible in the run.
    with ENV.patch(PEX_ROOT=pex_info.pex_root):
        if not execute:
            for location in _activate_pex(entry_point, pex_info, venv_dir=venv_dir):
                from pex.third_party import VendorImporter

                VendorImporter.install(
                    uninstallable=False, prefix="__pex__", path_items=["."], root=location
                )
            return

        interpreter_test = InterpreterTest(entry_point=entry_point, pex_info=pex_info)
        if not (ENV.PEX_UNZIP or ENV.PEX_TOOLS) and pex_info.venv:
            try:
                target = find_compatible_interpreter(interpreter_test=interpreter_test)
            except UnsatisfiableInterpreterConstraintsError as e:
                die(str(e))
            venv_pex = _bootstrap_venv(entry_point, interpreter=target)
            venv_pex.execv(*sys.argv[1:])
        else:
            maybe_reexec_pex(interpreter_test=interpreter_test)
            from . import pex

            pex.PEX(entry_point).execute()


def _activate_pex(
    entry_point,  # type: str
    pex_info,  # type: PexInfo
    venv_dir=None,  # type: Optional[str]
):
    # type: (...) -> Iterator[str]

    if pex_info.venv:
        for location in _activate_venv_dir(entry_point, venv_dir=venv_dir):
            yield location
        return

    from . import pex

    yield entry_point
    for distribution in pex.PEX(entry_point).activate():
        yield distribution.location


def _activate_venv_dir(
    entry_point,  # type: str
    venv_dir=None,  # type: Optional[str]
):
    # type: (...) -> Iterable[str]

    venv_python = None

    if venv_dir:
        python = os.path.join(venv_dir, "bin", "python")
        if os.path.exists(python):
            venv_python = python

    if not venv_python:
        venv_python = _bootstrap_venv(entry_point).python

    from pex.venv.virtualenv import Virtualenv

    venv = Virtualenv.enclosing(venv_python)
    if not venv:
        die("Failed to load virtualenv for interpreter at {path}.".format(path=venv_python))

    site_packages_dir = venv.site_packages_dir
    sys.path[:-1] = [site_packages_dir]
    import site

    site.addsitedir(site_packages_dir)
    yield site_packages_dir


def _bootstrap_venv(
    entry_point,  # type: str
    interpreter=None,  # type: Optional[PythonInterpreter]
):
    # type: (...) -> VenvPex

    from . import pex

    try:
        return ensure_venv(pex.PEX(entry_point, interpreter=interpreter))
    except ValueError as e:
        die(str(e))


# NB: This helper is used by third party libs - namely https://github.com/wickman/lambdex.
# TODO(John Sirois): Kill once https://github.com/wickman/lambdex/issues/5 is resolved.
def is_compressed(entry_point):
    # type: (str) -> bool
    return os.path.exists(entry_point) and not os.path.exists(
        os.path.join(entry_point, PexInfo.PATH)
    )


# NB: This helper is used by third party libs like https://github.com/wickman/lambdex and
# https://github.com/kwlzn/pyuwsgi_pex.
def bootstrap_pex_env(entry_point):
    # type: (str) -> None
    """Bootstrap the current runtime environment using a given pex."""
    pex_info = _bootstrap(entry_point)

    from .environment import PEXEnvironment

    PEXEnvironment.mount(entry_point, pex_info).activate()



# Copyright 2017 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

import itertools
import pkgutil
import re
from textwrap import dedent

import pytest

from pex.pep_425 import CompatibilityTags
from pex.platforms import Platform
from pex.third_party.packaging import tags

EXPECTED_BASE = [("py27", "none", "any"), ("py2", "none", "any")]


def test_platform():
    # type: () -> None
    assert Platform("linux-x86_64", "cp", "27", (2, 7), "mu") == Platform(
        "linux_x86_64", "cp", "27", (2, 7), "cp27mu"
    )
    assert Platform("linux-x86_64", "cp", "2.7", (2, 7), "mu") == Platform(
        "linux_x86_64", "cp", "2.7", (2, 7), "cp27mu"
    )

    assert str(Platform("linux-x86_64", "cp", "27", (2, 7), "m")) == "linux_x86_64-cp-27-cp27m"
    assert (
        str(Platform("linux-x86_64", "cp", "310", (3, 10), "cp310")) == "linux_x86_64-cp-310-cp310"
    )

    assert (
        str(Platform("linux-x86_64", "cp", "3.10", (3, 10), "cp310"))
        == "linux_x86_64-cp-3.10-cp310"
    )
    assert (
        str(Platform("linux-x86_64", "cp", "3.10.1", (3, 10, 1), "cp310"))
        == "linux_x86_64-cp-3.10.1-cp310"
    )


def test_platform_create():
    # type: () -> None
    assert Platform.create("linux-x86_64-cp-27-cp27mu") == Platform(
        "linux_x86_64", "cp", "27", (2, 7), "cp27mu"
    )
    assert Platform.create("linux-x86_64-cp-27-mu") == Platform(
        "linux_x86_64", "cp", "27", (2, 7), "cp27mu"
    )
    assert Platform.create("macosx-10.4-x86_64-cp-27-m") == Platform(
        "macosx_10_4_x86_64",
        "cp",
        "27",
        (2, 7),
        "cp27m",
    )


def assert_raises(platform, expected_cause):
    with pytest.raises(
        Platform.InvalidPlatformError,
        match=(
            r".*{literal}.*".format(
                literal=re.escape(
                    dedent(
                        """\
                        Not a valid platform specifier: {platform}
                        
                        {expected_cause}
                        """
                    ).format(platform=platform, expected_cause=expected_cause)
                )
            )
        ),
    ):
        Platform.create(platform)


def test_platform_create_bad_platform_missing_fields():
    # type: () -> None
    assert_raises(
        platform="linux_x86_64",
        expected_cause="There are missing platform fields. Expected 4 but given 1.",
    )


def test_platform_create_bad_platform_empty_fields():
    # type: () -> None
    assert_raises(
        platform="linux_x86_64--27-cp27mu",
        expected_cause="Platform specifiers cannot have blank fields. Given a blank impl.",
    )


def test_platform_create_bad_platform_bad_version():
    # type: () -> None
    assert_raises(
        platform="linux_x86_64-cp-2-cp27mu",
        expected_cause=(
            "The version field must either be a 2 or more digit digit major/minor version or else "
            "a component dotted version. Given: '2'"
        ),
    )

    assert_raises(
        platform="linux_x86_64-cp-XY-cp27mu",
        expected_cause="The version specified had non-integer components. Given: 'XY'",
    )

    assert_raises(
        platform="linux_x86_64-cp-2.-cp27mu",
        expected_cause="The version specified had non-integer components. Given: '2.'",
    )

    assert_raises(
        platform="linux_x86_64-cp-2.Y-cp27mu",
        expected_cause="The version specified had non-integer components. Given: '2.Y'",
    )


def test_platform_create_noop():
    # type: () -> None
    existing = Platform.create("linux-x86_64-cp-27-mu")
    assert Platform.create(existing) is existing


def test_platform_supported_tags():
    # type: () -> None
    platform = Platform.create("macosx-10.13-x86_64-cp-36-m")

    # A golden file test. This could break if we upgrade Pip and it upgrades packaging which, from
    # time to time, corrects omissions in tag sets.
    golden_tags = pkgutil.get_data(__name__, "data/platforms/macosx_10_13_x86_64-cp-36-m.tags.txt")
    assert golden_tags is not None
    assert (
        CompatibilityTags(
            itertools.chain.from_iterable(
                tags.parse_tag(tag)
                for tag in golden_tags.decode("utf-8").splitlines()
                if not tag.startswith("#")
            )
        )
        == platform.supported_tags()
    )


def test_platform_supported_tags_manylinux():
    # type: () -> None
    platform = Platform.create("linux-x86_64-cp-37-cp37m")
    tags = frozenset(platform.supported_tags())
    manylinux1_tags = frozenset(platform.supported_tags(manylinux="manylinux1"))
    manylinux2010_tags = frozenset(platform.supported_tags(manylinux="manylinux2010"))
    manylinux2014_tags = frozenset(platform.supported_tags(manylinux="manylinux2014"))
    assert manylinux2014_tags > manylinux2010_tags > manylinux1_tags > tags



# Copyright 2020 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

"""Constants to enable safe imports from the `typing` module.

This file needs to exist because Pex still supports running from Python 2. The `typing` stdlib
module is new to Python 3, so runtime imports of it will fail.

We don't want to vendor `typing` because it adds to PEX's bootstrap time and its bundle size. It's
also tricky to get working robustly.

Instead, we leverage the fact that MyPy ships with type stubs for Python 2. We also use the special
`TYPE_CHECKING` value, which in production always evaluates to `False`, but when running MyPy,
evaluates to True. This allows us to safely import `typing` without it ever being used in
production.

Note that we only use type comments, rather than normal annotations, which is what allows us to
never need `typing` at runtime except for `cast` and `overload` which have no-op runtime bindings
below.

To add type comments, use a conditional import like this:

    ```
    from pex.typing import TYPE_CHECKING

    if TYPE_CHECKING:
        from typing import Optional, ...
    ```
"""

from __future__ import absolute_import, print_function

import sys

TYPE_CHECKING = False

# Unlike most type-hints, `cast` and `overload` get used at runtime. We define no-op versions for
# runtime use.
if TYPE_CHECKING:
    from typing import Any
    from typing import Generic as Generic
    from typing import cast as cast
    from typing import overload as overload

    if sys.version_info[:2] >= (3, 8):
        from typing import Literal as Literal
    else:
        from typing_extensions import Literal as Literal
else:

    def cast(_type, value):
        return value

    def overload(_func):
        def _never_called_since_structurally_shadowed(*_args, **_kwargs):
            raise NotImplementedError(
                "You should not call an overloaded function. A series of @overload-decorated "
                "functions outside a stub module should always be followed by an implementation "
                "that is not @overload-ed."
            )

        return _never_called_since_structurally_shadowed

    class _Generic(type):
        def __getitem__(cls, type_var):
            return cls

    if sys.version_info[0] == 2:

        class Generic(object):
            __metaclass__ = _Generic

    else:
        eval(compile("class Generic(object, metaclass=_Generic): pass", "<Generic>", "exec"))

    del _Generic

    Literal = {}



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.resolve.locked_resolve import (
    FileArtifact,
    LocalProjectArtifact,
    LockedRequirement,
    LockedResolve,
    VCSArtifact,
)
from pex.sorted_tuple import SortedTuple
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Union

    import attr  # vendor:skip
else:
    from pex.third_party import attr


def normalize_artifact(
    artifact,  # type: Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    skip_urls=False,  # type: bool
    skip_verified=False,  # type: bool
):
    # type: (...) -> Union[FileArtifact, LocalProjectArtifact, VCSArtifact]
    return attr.evolve(
        artifact,
        url="" if skip_urls else artifact.url,
        verified=False if skip_verified else artifact.verified,
    )


def normalize_locked_requirement(
    locked_req,  # type: LockedRequirement
    skip_additional_artifacts=False,  # type: bool
    skip_urls=False,  # type: bool
    skip_verified=False,  # type: bool
):
    # type: (...) -> LockedRequirement
    return attr.evolve(
        locked_req,
        artifact=normalize_artifact(
            locked_req.artifact, skip_urls=skip_urls, skip_verified=skip_verified
        ),
        additional_artifacts=()
        if skip_additional_artifacts
        else SortedTuple(
            normalize_artifact(a, skip_urls=skip_urls, skip_verified=skip_verified)
            for a in locked_req.additional_artifacts
        ),
    )


def normalize_locked_resolve(
    lock,  # type: LockedResolve
    skip_additional_artifacts=False,  # type: bool
    skip_urls=False,  # type: bool
    skip_verified=False,  # type: bool
):
    # type: (...) -> LockedResolve
    return attr.evolve(
        lock,
        locked_requirements=SortedTuple(
            normalize_locked_requirement(
                locked_req,
                skip_additional_artifacts=skip_additional_artifacts,
                skip_urls=skip_urls,
                skip_verified=skip_verified,
            )
            for locked_req in lock.locked_requirements
        ),
    )



# Copyright 2022 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from pex.dist_metadata import Requirement
from pex.enum import Enum
from pex.pep_440 import Version
from pex.targets import LocalInterpreter, Target
from pex.third_party.packaging.specifiers import SpecifierSet
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Iterable, Optional, Tuple


class PipVersionValue(Enum.Value):
    def __init__(
        self,
        version,  # type: str
        name=None,  # type: Optional[str]
        requirement=None,  # type: Optional[str]
        setuptools_version=None,  # type: Optional[str]
        wheel_version=None,  # type: Optional[str]
        requires_python=None,  # type: Optional[str]
    ):
        # type: (...) -> None
        super(PipVersionValue, self).__init__(name or version)

        def to_requirement(
            project_name,  # type: str
            project_version=None,  # type: Optional[str]
        ):
            # type: (...) -> str
            return (
                "{project_name}=={project_version}".format(
                    project_name=project_name, project_version=project_version
                )
                if project_version
                else project_name
            )

        self.version = Version(version)
        self.requirement = requirement or to_requirement("pip", version)
        self.setuptools_requirement = to_requirement("setuptools", setuptools_version)
        self.wheel_requirement = to_requirement("wheel", wheel_version)
        self.requires_python = SpecifierSet(requires_python) if requires_python else None

    @property
    def requirements(self):
        # type: () -> Iterable[str]
        return self.requirement, self.setuptools_requirement, self.wheel_requirement

    def requires_python_applies(self, target):
        # type: (Target) -> bool
        if not self.requires_python:
            return True

        return LocalInterpreter.create(target.get_interpreter()).requires_python_applies(
            requires_python=self.requires_python,
            source=Requirement.parse(self.requirement),
        )


class LatestPipVersion(object):
    def __get__(self, obj, objtype=None):
        if not hasattr(self, "_latest"):
            self._latest = max(PipVersionValue._iter_values(), key=lambda pv: pv.version)
        return self._latest


class PipVersion(Enum["PipVersionValue"]):
    @classmethod
    def values(cls):
        # type: () -> Tuple[PipVersionValue, ...]
        if cls._values is None:
            cls._values = tuple(PipVersionValue._iter_values())
        return cls._values

    v20_3_4_patched = PipVersionValue(
        name="20.3.4-patched",
        version="20.3.4+patched",
        requirement=(
            "pip @ git+https://github.com/pantsbuild/pip@386a54f097ece66775d0c7f34fd29bb596c6b0be"
        ),
    )

    # TODO(John Sirois): Expose setuptools and wheel version flags - these don't affect
    #  Pex; so we should allow folks to experiment with upgrade easily:
    #  https://github.com/pantsbuild/pex/issues/1895

    v22_2_2 = PipVersionValue(
        version="22.2.2",
        setuptools_version="65.3.0",
        wheel_version="0.37.1",
        requires_python=">=3.7",
    )

    v22_3 = PipVersionValue(
        version="22.3",
        setuptools_version="65.5.0",
        wheel_version="0.37.1",
        requires_python=">=3.7",
    )

    v22_3_1 = PipVersionValue(
        version="22.3.1",
        setuptools_version="65.5.1",
        wheel_version="0.37.1",
        requires_python=">=3.7",
    )

    v23_0 = PipVersionValue(
        version="23.0",
        setuptools_version="67.2.0",
        wheel_version="0.38.4",
        requires_python=">=3.7",
    )

    v23_0_1 = PipVersionValue(
        version="23.0.1",
        setuptools_version="67.4.0",
        wheel_version="0.38.4",
        requires_python=">=3.7",
    )

    v23_1 = PipVersionValue(
        version="23.1",
        setuptools_version="67.6.1",
        wheel_version="0.40.0",
        requires_python=">=3.7",
    )

    v23_1_1 = PipVersionValue(
        version="23.1.1",
        setuptools_version="67.7.1",
        wheel_version="0.40.0",
        requires_python=">=3.7",
    )

    v23_1_2 = PipVersionValue(
        version="23.1.2",
        setuptools_version="67.7.2",
        wheel_version="0.40.0",
        requires_python=">=3.7",
    )

    VENDORED = v20_3_4_patched
    LATEST = LatestPipVersion()



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import print_function

import os
from textwrap import dedent

from pex.testing import run_pex_command
from pex.typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


def test_undeclared_setuptools_import_on_pex_path(tmpdir):
    # type: (Any) -> None
    """Test that packages which access pkg_resources at import time can be found with pkg_resources.

    See https://github.com/pantsbuild/pex/issues/729 for context. We warn when a package accesses
    pkg_resources without declaring it in install_requires, but we also want to check that those
    packages can be accessed successfully via the PEX_PATH.
    """
    setuptools_pex = os.path.join(str(tmpdir), "setuptools.pex")
    # NB: the specific setuptools version does not necessarily matter. We only pin the version to
    # avoid a future version of setuptools potentially fixing this issue and then us no longer
    # checking that Pex is behaving properly for older setuptools versions.
    run_pex_command(["setuptools==40.6.3", "-o", setuptools_pex]).assert_success()

    # We constrain google-crc32c to avoid compilation errors in CI for PyPy which has no published
    # wheels.
    #
    # Via google-resumable-media>=0.3.1 which nets 2.0.0, via google-crc32c>=0.3.1 which nets 1.1.3
    # which fails to compile with:
    #
    # gcc -pthread -DNDEBUG -O2 -fPIC -I/opt/hostedtoolcache/PyPy/3.6.12/x64/include -c src/google_crc32c/_crc32c.c -o build/temp.linux-x86_64-3.6/src/google_crc32c/_crc32c.o
    # src/google_crc32c/_crc32c.c:3:10: fatal error: crc32c/crc32c.h: No such file or directory
    #     3 | #include <crc32c/crc32c.h>
    #       |          ^~~~~~~~~~~~~~~~~
    # compilation terminated.
    constraints = os.path.join(str(tmpdir), "constraints.txt")
    with open(constraints, "w") as fp:
        print("google-api-core==1.32.0", file=fp)
        print("googleapis-common-protos<=1.56.4", file=fp)
        print("google-crc32c==1.1.2", file=fp)
        print("protobuf<=3.17.3", file=fp)

    bigquery_pex = os.path.join(str(tmpdir), "bigquery.pex")
    run_pex_command(
        args=["google-cloud-bigquery==1.10.0", "--constraints", constraints, "-o", bigquery_pex]
    ).assert_success()

    src_dir = os.path.join(str(tmpdir), "src")
    os.mkdir(src_dir)

    src_file = os.path.join(src_dir, "execute_import.py")
    with open(src_file, "w") as fp:
        fp.write(
            dedent(
                """\
                from google.cloud import bigquery

                print('bigquery version: {}'.format(bigquery.__version__))
                """
            )
        )

    res = run_pex_command(
        [
            "--pex-path={}".format(os.pathsep.join([setuptools_pex, bigquery_pex])),
            "-D",
            src_dir,
            "--entry-point",
            "execute_import",
        ]
    )
    res.assert_success()
    assert res.output.strip() == "bigquery version: 1.10.0"



# Copyright 2021 Pants project contributors (see CONTRIBUTORS.md).
# Licensed under the Apache License, Version 2.0 (see LICENSE).

from __future__ import absolute_import

from abc import abstractmethod

from pex.commands.command import Command
from pex.pex import PEX
from pex.result import Result


class PEXCommand(Command):
    @abstractmethod
    def run(self, pex):
        # type: (PEX) -> Result
        raise NotImplementedError()
